/*	$Id: adb.c,v 1.12 2003/01/17 17:40:16 tsubai Exp $	*/

/*-
 * Copyright (C) 1994	Bradley A. Grantham
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bradley A. Grantham.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/fcntl.h>
#include <sys/poll.h>
#include <sys/proc.h>
#include <sys/select.h>
#include <sys/signalvar.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/pio.h>

#include <macppc/dev/adbvar.h>
#include <macppc/dev/akbdvar.h>
#include <macppc/dev/pm_direct.h>
#include <macppc/dev/viareg.h>

#include <dev/ofw/openfirm.h>

#include "aed.h"

struct adb_knowndev {
	u_char addr;
	u_char handler_id;
	char *desc;
};

#define ADBDEVS_DATA
#include <macppc/dev/adbdevs.h>

/*
 * Function declarations.
 */
int adbmatch __P((struct device *, struct cfdata *, void *));
void adbattach __P((struct device *, struct device *, void *));
int adbprint __P((void *, const char *));
void adb_powerhook(int, void *);
static void install_NMI(void);
#ifdef DDB
static int nmi(void *);
#endif

/*
 * Global variables.
 */
int     adb_polling = 0;	/* Are we polling?  (Debugger mode) */
int     adb_initted = 0;	/* adb_init() has completed successfully */
#ifdef ADB_DEBUG
int	adb_debug = 0;		/* Output debugging messages */
#endif /* ADB_DEBUG */

/*
 * Driver definition.
 */
struct cfattach adb_ca = {
	"adb", {}, sizeof(struct adb_softc), adbmatch, adbattach
};

extern int adbHardware;

int
adbmatch(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (ca->ca_nreg < 8)
		return 0;

	if (ca->ca_nintr < 4)
		return 0;

	if (strcmp(ca->ca_name, "via-cuda") == 0)
		return 1;

	if (strcmp(ca->ca_name, "via-pmu") == 0)
		return 1;

	return 0;
}

void
adbattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct adb_softc *sc = (struct adb_softc *)self;
	struct confargs *ca = aux;
	int irq = ca->ca_intr[0];
	int node;
	ADBDataBlock adbdata;
	struct adb_attach_args aa;
	int totaladbs;
	int adbindex, adbaddr;

	extern volatile u_char *Via1Base;

	ca->ca_reg[0] += ca->ca_baseaddr;

	sc->sc_regbase = mapiodev(ca->ca_reg[0], ca->ca_reg[1]);
	Via1Base = sc->sc_regbase;

	if (strcmp(ca->ca_name, "via-cuda") == 0)
		adbHardware = ADB_HW_CUDA;
	else if (strcmp(ca->ca_name, "via-pmu") == 0)
		adbHardware = ADB_HW_PB;

	node = getnodebyname(OF_parent(ca->ca_node), "extint-gpio1");
	if (node)
		OF_getprop(node, "interrupts", &irq, 4);

	printf(" irq %d: ", irq);

	adb_polling = 1;
	ADBReInit();

	intr_establish(irq, IST_LEVEL, IPL_HIGH, adb_intr, sc);

#ifdef ADB_DEBUG
	if (adb_debug)
		printf("adb: done with ADBReInit\n");
#endif
	totaladbs = CountADBs();

	printf("%d targets\n", totaladbs);

#if NAED > 0
	/* ADB event device for compatibility */
	aa.origaddr = 0;
	aa.adbaddr = 0;
	aa.handler_id = 0;
	(void)config_found(self, &aa, adbprint);
#endif

	/* for each ADB device */
	for (adbindex = 1; adbindex <= totaladbs; adbindex++) {
		/* Get the ADB information */
		adbaddr = GetIndADB(&adbdata, adbindex);

		aa.origaddr = adbdata.origADBAddr;
		aa.adbaddr = adbaddr;
		aa.handler_id = adbdata.devType;

		(void)config_found(self, &aa, adbprint);
	}

	if (adbHardware == ADB_HW_CUDA)
		adb_cuda_autopoll();
	adb_polling = 0;

	install_NMI();
	powerhook_establish(adb_powerhook, sc);
}

#define NMI_INTSRC 0x800506e0
#define NMI_LEVEL  0x00004000
#define NMI_ENABLE 0x00000800

void
install_NMI()
{
#ifdef DDB
	int node;
	u_int x, reg[2];

	if (OF_finddevice("/uni-n") == -1)
		return;
	node = getnodebyname(0, "programmer-switch");
	if (node == 0)
		return;
	if (OF_getprop(node, "interrupts", reg, sizeof reg) != 8)
		return;

	x = in32(NMI_INTSRC);

	/* Use level trigger. */
	x |= NMI_LEVEL;
	out32(NMI_INTSRC, x);

	/* Enable NMI. */
	x |= NMI_ENABLE;
	out32(NMI_INTSRC, x);

	intr_establish(reg[0], IST_LEVEL, IPL_HIGH, nmi, NULL);
#endif
}

#ifdef DDB
int
nmi(v)
	void *v;
{
	Debugger();
	return 1;
}
#endif

int
adbprint(args, name)
	void *args;
	const char *name;
{
	struct adb_attach_args *aa = args;
	int id = aa->handler_id;
	int rv = UNCONF;

	if (name) {	/* no configured device matched */
		rv = UNSUPP; /* most ADB device types are unsupported */

		/* print out what kind of ADB device we have found */
		printf("%s addr %d: ", name, aa->adbaddr);
		switch(aa->origaddr) {
#ifdef DIAGNOSTIC
		case 0:
			printf("ADB event device");
			rv = UNCONF;
			break;
		case ADBADDR_SECURE:
			printf("security dongle (%d)", id);
			break;
#endif
		case ADBADDR_MAP:
			printf("mapped device (%d)", id);
			rv = UNCONF;
			break;
		case ADBADDR_REL:
			printf("relative positioning device (%d)", id);
			rv = UNCONF;
			break;
#ifdef DIAGNOSTIC
		case ADBADDR_ABS:
			switch (id) {
			case ADB_ARTPAD:
				printf("WACOM ArtPad II");
				break;
			default:
				printf("absolute positioning device (%d)", id);
				break;
			}
			break;
		case ADBADDR_DATATX:
			printf("data transfer device (modem?) (%d)", id);
			break;
		case ADBADDR_MISC:
			switch (id) {
			case ADB_POWERKEY:
				printf("Sophisticated Circuits PowerKey");
				break;
			case ADB_STUDIODISP:
				printf("Apple Studio Display's control");
				break;
			default:
				printf("misc. device (remote control?) (%d)",
				    id);
				break;
			}
			break;
		default:
			printf("unknown type device, (handler %d)", id);
			break;
#endif /* DIAGNOSTIC */
		}
	} else		/* a device matched and was configured */
		printf(" addr %d: ", aa->adbaddr);

	return rv;
}

void
adb_devinfo(addr, id, desc)
	int addr, id;
	char *desc;
{
	const struct adb_knowndev *dp = adb_knowndevs;

	while (dp->desc) {
		if (dp->addr == addr && dp->handler_id == id) {
			strcpy(desc, dp->desc);
			return;
		}
		dp++;
	}

	switch (addr) {

	case ADBADDR_KBD:
		sprintf(desc, "mapped device (%d)", id);
		return;

	case ADBADDR_MS:
		sprintf(desc, "relative positioning device (mouse?) (%d)", id);
		return;
	}

	*desc = 0;
}

void
adb_powerhook(why, v)
	int why;
	void *v;
{
	if (adbHardware == ADB_HW_PB)
		pm_powerhook(why, v);
}
