/*	$Id: hwsensor.c,v 1.15 2003/10/18 17:40:42 tsubai Exp $	*/

/*-
 * Copyright (c) 2003 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * DS1775 manual:
 * http://pdfserv.maxim-ic.com/en/ds/DS1775-DS1775R.pdf
 *
 * ADT7460:
 * http://www.analog.com/UploadedFiles/Data_Sheets/3686221171167ADT7460_b.pdf
 *
 * ADM1031:
 * http://www.analog.com/UploadedFiles/Data_Sheets/27250527ADM1031_a.pdf
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <dev/ofw/openfirm.h>
#include <dev/sysmon/sysmonvar.h>

#include <machine/autoconf.h>
#include <machine/cpufunc.h>
#include <machine/uni-n.h>

enum {
	TEMP1 = 0,
	TEMP2,
	TEMP3,
	FAN1,
	FAN2,
	BUSCLOCK,
	CPUCLOCK,
	NSENSORS
};

struct sysmon_cookie {
	struct hwsensor_softc *softc;
	int sensor;
};

struct hwsensor_softc {
	struct device sc_dev;
	void *sc_i2c;
	int sc_ui2c;
	u_int sc_busfreq;

	struct sysmon_envsys sc_sysmon[NSENSORS];
	struct envsys_tre_data sc_data[NSENSORS];
	struct envsys_basic_info sc_info[NSENSORS];
	struct sysmon_cookie sc_cookie[NSENSORS];
};

int hwsensor_match(struct device *, struct cfdata *, void *);
void hwsensor_attach(struct device *, struct device *, void *);
int hwsensor_sysmon_attach(struct hwsensor_softc *);
int hwsensor_streinfo(struct sysmon_envsys *, struct envsys_basic_info *);
static int gtredata_adt7460(struct sysmon_envsys *, struct envsys_tre_data *);
static int gtredata_ds1775(struct sysmon_envsys *, struct envsys_tre_data *);

struct cfattach hwsensor_ca = {
	"hwsensor", {}, sizeof(struct hwsensor_softc),
	hwsensor_match, hwsensor_attach
};

int
hwsensor_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "uni-n") != 0)
		return 0;

	if (strncmp(macppc_model, "PowerBook3,", 11) == 0)
		return 1;
	if (strncmp(macppc_model, "PowerBook4,", 11) == 0)
		return 1;
	if (strncmp(macppc_model, "PowerBook5,3", 12) == 0)
		return 1;
	if (strncmp(macppc_model, "RackMac1,", 9) == 0)
		return 1;

	return 0;
}

void
hwsensor_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct hwsensor_softc *sc = (void *)self;
	int node;
	char type[32];

	/* Initialize i2c interface. */
	sc->sc_ui2c = OF_finddevice("ui2c");
	if (sc->sc_ui2c == -1) {
		printf(": cannot find ui2c\n");
		return;
	}

	/* Get bus clock frequency. */
	node = OF_finddevice("/cpus");
	if (node == -1)
		node = OF_finddevice("/");
	if (node == -1)
		panic("no root node?");
	for (node = OF_child(node); node; node = OF_peer(node)) {
		bzero(type, sizeof type);
		OF_getprop(node, "device_type", type, sizeof type);
		if (strcmp(type, "cpu") == 0)
			goto found;
	}
	printf(": no cpu node?");
	return;

found:
	OF_getprop(node, "bus-frequency", &sc->sc_busfreq, 4);

	hwsensor_sysmon_attach(sc);
}

static struct envsys_range hwsensor_ranges[] = {
	{ 1, 0, ENVSYS_STEMP },
	{ 1, 0, ENVSYS_SFANRPM },
	{ 1, 0, ENVSYS_SVOLTS_AC },
	{ 1, 0, ENVSYS_SVOLTS_DC },
	{ 1, 0, ENVSYS_SOHMS },
	{ 1, 0, ENVSYS_SWATTS },
	{ 1, 0, ENVSYS_SAMPS },
};

#define DS1775_ADDR 0x92
#define ADT7460_ADDR 0x15c

int
hwsensor_sysmon_attach(sc)
	struct hwsensor_softc *sc;
{
	int i;
	int fan, addr, adt7460 = 0;
	char compat[32];

	addr = DS1775_ADDR;

	fan = OF_finddevice("/uni-n/i2c/fan");
	if (fan != -1) {
		bzero(compat, sizeof compat);
		OF_getprop(fan, "compatible", compat, sizeof compat);
		if (strcmp(compat, "adt7460") == 0) {
			adt7460 = 1;
			if (OF_getprop(fan, "reg", &addr, 4) != 4)
				addr = ADT7460_ADDR;
		}
	}
	if (adt7460)
		printf(": adt7460 addr 0x%x\n", addr);
	else
		printf(": ds1775 addr 0x%x\n", addr);
	sc->sc_i2c = ki2c_open(sc->sc_ui2c, addr, I2C_STDMODE);

	/* Initialize sensors. */
	strcpy(sc->sc_info[TEMP1].desc, "temp1");
	strcpy(sc->sc_info[TEMP2].desc, "temp2");
	strcpy(sc->sc_info[TEMP3].desc, "temp3");
	strcpy(sc->sc_info[FAN1].desc, "fan1");
	strcpy(sc->sc_info[FAN2].desc, "fan2");
	strcpy(sc->sc_info[BUSCLOCK].desc, "bus-clock");
	strcpy(sc->sc_info[CPUCLOCK].desc, "cpu-clock");

	sc->sc_data[TEMP1].units = ENVSYS_STEMP;
	sc->sc_data[TEMP2].units = ENVSYS_STEMP;
	sc->sc_data[TEMP3].units = ENVSYS_STEMP;
	sc->sc_data[FAN1].units = ENVSYS_SFANRPM;
	sc->sc_data[FAN2].units = ENVSYS_SFANRPM;
	sc->sc_data[BUSCLOCK].units = ENVSYS_INTEGER;
	sc->sc_data[CPUCLOCK].units = ENVSYS_INTEGER;

	sc->sc_info[TEMP1].validflags = ENVSYS_FVALID;
	if (adt7460) {
		sc->sc_info[TEMP2].validflags = ENVSYS_FVALID;
		sc->sc_info[TEMP3].validflags = ENVSYS_FVALID;
		sc->sc_info[FAN1].validflags = ENVSYS_FVALID;
		sc->sc_info[FAN2].validflags = ENVSYS_FVALID;
	}

	switch (mfpvr() >> 16) {
	case MPC7450:
	case MPC7455:
	case MPC7457:
		sc->sc_info[BUSCLOCK].validflags = ENVSYS_FVALID;
		sc->sc_info[CPUCLOCK].validflags = ENVSYS_FVALID;
	}

	for (i = 0; i < NSENSORS; i++) {
		if ((sc->sc_info[i].validflags & ENVSYS_FVALID) == 0)
			continue;

		sc->sc_data[i].sensor = 0;
		sc->sc_data[i].validflags = ENVSYS_FVALID | ENVSYS_FCURVALID;
		sc->sc_data[i].warnflags = ENVSYS_WARN_OK;
		sc->sc_info[i].sensor = 0;
		sc->sc_info[i].units = sc->sc_data[i].units;

		sc->sc_sysmon[i].sme_nsensors = 1;
		sc->sc_sysmon[i].sme_envsys_version = 1000;
		sc->sc_sysmon[i].sme_ranges = hwsensor_ranges;
		sc->sc_sysmon[i].sme_sensor_data = &sc->sc_data[i];
		sc->sc_sysmon[i].sme_sensor_info = &sc->sc_info[i];
		if (adt7460)
			sc->sc_sysmon[i].sme_gtredata = gtredata_adt7460;
		else
			sc->sc_sysmon[i].sme_gtredata = gtredata_ds1775;
		sc->sc_sysmon[i].sme_streinfo = hwsensor_streinfo;
		sc->sc_sysmon[i].sme_cookie = &sc->sc_cookie[i];
		sc->sc_cookie[i].softc = sc;
		sc->sc_cookie[i].sensor = i;
		sysmon_envsys_register(&sc->sc_sysmon[i]);
	}
	return 0;
}

int
gtredata_ds1775(sme, tred)
	struct sysmon_envsys *sme;
	struct envsys_tre_data *tred;
{
	struct sysmon_cookie *c = sme->sme_cookie;
	struct hwsensor_softc *sc = c->softc;
	int sensor = c->sensor;
	int rv;
	u_char data[32];

	if ((sc->sc_info[sensor].validflags & ENVSYS_FVALID) == 0)
		return 0;

	switch (sensor) {

	case TEMP1:
		bzero(data, sizeof data);
		rv = ki2c_write(sc->sc_i2c, data, 1);
		rv = ki2c_read(sc->sc_i2c, data, 2);
		sc->sc_data[sensor].cur.data_s =
		    data[0] * 1000000 + data[1] * 1000000 / 256 + 273150000;
		break;

	case BUSCLOCK:
		sc->sc_data[sensor].cur.data_us = sc->sc_busfreq;
		break;

	case CPUCLOCK:
		sc->sc_data[sensor].cur.data_us =
		    mpc7450_corefreq(sc->sc_busfreq);
		break;
	}

	*tred = sc->sc_data[sensor];
	return 0;
}

static int adt7460_read(struct hwsensor_softc *, int);

int
adt7460_read(sc, reg)
	struct hwsensor_softc *sc;
	int reg;
{
	u_char data[32];

	data[0] = reg;
	if (ki2c_write(sc->sc_i2c, data, 1))
		return -1;
	if (ki2c_read(sc->sc_i2c, data, 1))
		return -1;
	return data[0];
}

int
gtredata_adt7460(sme, tred)
	struct sysmon_envsys *sme;
	struct envsys_tre_data *tred;
{
	struct sysmon_cookie *c = sme->sme_cookie;
	struct hwsensor_softc *sc = c->softc;
	int sensor = c->sensor;
	int t, rpm;

	if ((sc->sc_info[sensor].validflags & ENVSYS_FVALID) == 0)
		return 0;

	switch (sensor) {
#if 0
	case P25V:
		/* 2.5V */
		t = adt7460_read(sc, 0x20);
		printf("0x20 0x%x\n", t);

	case VCC:
		/* Vcc (3.3V) */
		t = adt7460_read(sc, 0x22);
		printf("0x22 0x%x\n", t);
#endif

	case TEMP1:
		/* Local Temp */
		t = adt7460_read(sc, 0x26);
		sc->sc_data[sensor].cur.data_s = t * 1000000 + 273150000;
		break;

	case TEMP2:
		/* Remote1 Temp */
		t = adt7460_read(sc, 0x25);
		sc->sc_data[sensor].cur.data_s = t * 1000000 + 273150000;
		break;

	case TEMP3:
		/* Temote2 Temp */
		t = adt7460_read(sc, 0x27);
		sc->sc_data[sensor].cur.data_s = t * 1000000 + 273150000;
		break;

	case FAN1:
		/* TACH1 */
		t  = adt7460_read(sc, 0x28);
		t += adt7460_read(sc, 0x29) << 8;
		rpm = 5400000 / t;
		if (t == 0xffff)	/* Not running */
			rpm = 0;
		sc->sc_data[sensor].cur.data_us = rpm;
		break;

	case FAN2:
		/* TACH2 */
		t  = adt7460_read(sc, 0x2a);
		t += adt7460_read(sc, 0x2b) << 8;
		rpm = 5400000 / t;
		if (t == 0xffff)	/* Not running */
			rpm = 0;
		sc->sc_data[sensor].cur.data_us = rpm;
		break;

	case BUSCLOCK:
		sc->sc_data[sensor].cur.data_us = sc->sc_busfreq;
		break;

	case CPUCLOCK:
		sc->sc_data[sensor].cur.data_us =
		    mpc7450_corefreq(sc->sc_busfreq);
		break;
	}

	*tred = sc->sc_data[sensor];
	return 0;
}

int
hwsensor_streinfo(sme, binfo)
	struct sysmon_envsys *sme;
	struct envsys_basic_info *binfo;
{
	printf("hwsensor_streinfo sensor = %d\n", binfo->sensor);

	binfo->validflags = 0;
	return 0;
}

/*
 Xserve:
/uni-n@f8000000/i2c@f8001000/temp-monitor@92   
name                    temp-monitor
device_type             ds1775
compatible              ds1775
reg                     00000092

 PBG4 1.3G:
/uni-n@f8000000/i2c@f8001000/fan@15c
name                    fan
device_type             adt7460
compatible              adt7460
reg                     0000015c

*/
