/*	$Id: if_gm.c,v 1.46 2003/10/29 13:12:00 tsubai Exp $	*/

/*-
 * Copyright (c) 2000,2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "bpfilter.h"

#include <sys/param.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/systm.h>
#include <sys/callout.h>

#include <uvm/uvm_extern.h>

#include <net/if.h>
#include <net/if_ether.h>
#include <net/if_media.h>

#include <netinet/in.h>			/* XXX for struct ip */
#include <netinet/in_systm.h>		/* XXX for struct ip */
#include <netinet/ip.h>			/* XXX for struct ip */

#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#include <dev/mii/mii.h>
#include <dev/mii/miivar.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcidevs.h>

#include <dev/ofw/openfirm.h>
#include <macppc/dev/if_gmreg.h>
#include <machine/pio.h>

#define NTXBUF 64	/* 32 - 8192, power-of-2. */
#define NRXBUF 64	/* 32 - 8192, power-of-2. */

#define sc_if sc_ethercom.ec_if

struct gmac_softc {
	struct device sc_dev;
	struct ethercom sc_ethercom;
	vaddr_t sc_reg;
	struct gmac_dma *sc_txlist;
	struct gmac_dma *sc_rxlist;
	int sc_txnext;
	int sc_txfree;
	int sc_rxlast;
	struct mbuf *sc_txbuf[NTXBUF];
	struct mbuf *sc_rxbuf[NRXBUF];
	struct mii_data sc_mii;
	struct callout sc_tick_ch;
	char sc_laddr[6];
};

int gmac_match __P((struct device *, struct cfdata *, void *));
void gmac_attach __P((struct device *, struct device *, void *));

static inline u_int gmac_read_reg __P((struct gmac_softc *, int));
static inline void gmac_write_reg __P((struct gmac_softc *, int, u_int));

static inline void gmac_start_txdma __P((struct gmac_softc *));
static inline void gmac_start_rxdma __P((struct gmac_softc *));
static inline void gmac_stop_txdma __P((struct gmac_softc *));
static inline void gmac_stop_rxdma __P((struct gmac_softc *));

int gmac_intr __P((void *));
void gmac_tint __P((struct gmac_softc *));
void gmac_rint __P((struct gmac_softc *));
void gmac_rxhwsum(struct mbuf *, u_int);
struct mbuf *gmac_get __P((struct gmac_softc *));
void gmac_start __P((struct ifnet *));
void gmac_stop __P((struct gmac_softc *));
void gmac_reset __P((struct gmac_softc *));
int gmac_init __P((struct ifnet *));
void gmac_init_mac __P((struct gmac_softc *));
void gmac_setladrf __P((struct gmac_softc *));

int gmac_ioctl __P((struct ifnet *, u_long, caddr_t));
void gmac_watchdog __P((struct ifnet *));

int gmac_mediachange __P((struct ifnet *));
void gmac_mediastatus __P((struct ifnet *, struct ifmediareq *));
int gmac_mii_readreg __P((struct device *, int, int));
void gmac_mii_writereg __P((struct device *, int, int, int));
void gmac_mii_statchg __P((struct device *));
void gmac_tick __P((void *));
void gmac_powerhook(int, void *);

struct cfattach gm_ca = {
	"gm", {}, sizeof(struct gmac_softc), gmac_match, gmac_attach
};

int
gmac_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct pci_attach_args *pa = aux;

	if (PCI_VENDOR(pa->pa_id) == PCI_VENDOR_APPLE &&
	    (PCI_PRODUCT(pa->pa_id) == PCI_PRODUCT_APPLE_GMAC ||
	     PCI_PRODUCT(pa->pa_id) == PCI_PRODUCT_APPLE_GMAC2 ||
	     PCI_PRODUCT(pa->pa_id) == PCI_PRODUCT_APPLE_GMAC3))
		return 2;

	return 0;
}

void
gmac_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct gmac_softc *sc = (void *)self;
	struct pci_attach_args *pa = aux;
	struct ifnet *ifp = &sc->sc_if;
	struct mii_data *mii = &sc->sc_mii;
	pci_intr_handle_t ih;
	const char *intrstr = NULL;
	int node, phy, i;
	int flags = 0;
	char *p;
	struct gmac_dma *dp;
	u_int32_t reg[10];
	u_char laddr[6];

	node = getnodebypci(pa->pa_pc, pa->pa_tag);
	if (node == 0) {
		printf(": cannot find gmac node\n");
		return;
	}

	OF_getprop(node, "local-mac-address", laddr, sizeof laddr);
	OF_getprop(node, "assigned-addresses", reg, sizeof reg);

	phy = getnodebyname(node, "ethernet-phy");
	if (phy) {
		char compat[16];

		bzero(compat, sizeof compat);
		OF_getprop(phy, "compatible", compat, sizeof compat);
		if (strcmp(compat, "B5201") == 0 ||
		    strcmp(compat, "B5221") == 0 ||
		    strcmp(compat, "B5411") == 0 ||
		    strcmp(compat, "B5421") == 0)
			flags |= MIIF_FORCEANEG;
	} else {
		/* XXX To be safe. */
		flags |= MIIF_FORCEANEG;
	}

	bcopy(laddr, sc->sc_laddr, sizeof laddr);
	sc->sc_reg = reg[2];

	if (pci_intr_map(pa, &ih)) {
		printf(": unable to map interrupt\n");
		return;
	}
	intrstr = pci_intr_string(pa->pa_pc, ih);

	if (pci_intr_establish(pa->pa_pc, ih, IPL_NET, gmac_intr, sc) == NULL) {
		printf(": unable to establish interrupt");
		if (intrstr)
			printf(" at %s", intrstr);
		printf("\n");
		return;
	}

	/* Setup DMA descriptors. */
	p = malloc((NTXBUF + NRXBUF) * sizeof(struct gmac_dma) + 4096,
	    M_DEVBUF, M_NOWAIT | M_ZERO);
	if (p == NULL) {
		printf(": cannot malloc buffers\n");
		return;
	}
	p = (void *)roundup((vaddr_t)p, 2048);

	sc->sc_rxlist = (void *)p;
	p += NRXBUF * sizeof(struct gmac_dma);
	p = (void *)roundup((vaddr_t)p, 2048);

	sc->sc_txlist = (void *)p;

	/* Initialize receive ring. */
	dp = sc->sc_rxlist;
	for (i = 0; i < NRXBUF; i++) {
		struct mbuf *m;

		m = gmac_get(sc);
		if (m == NULL) {
			printf(": cannot alloc mbufs\n");
			return;
		}
		sc->sc_rxbuf[i] = m;
#ifdef PMAP_UNMAP_POOLPAGE
		dp->address = htole32(PMAP_UNMAP_POOLPAGE(mtod(m, vaddr_t)));
#else
		dp->address = htole32(vtophys(mtod(m, vaddr_t)));
#endif
		dp->cmd = htole32(GMAC_OWN);
		dp++;
	}

	printf(": Ethernet address %s\n", ether_sprintf(laddr));
	printf("%s: interrupting at %s\n", sc->sc_dev.dv_xname, intrstr);

	callout_init(&sc->sc_tick_ch);

	gmac_reset(sc);
	gmac_init_mac(sc);

	bcopy(sc->sc_dev.dv_xname, ifp->if_xname, IFNAMSIZ);
	ifp->if_softc = sc;
	ifp->if_ioctl = gmac_ioctl;
	ifp->if_init = gmac_init;
	ifp->if_start = gmac_start;
	ifp->if_watchdog = gmac_watchdog;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	IFQ_SET_READY(&ifp->if_snd);

	/* We can perform TCPv4 checksum. */
	ifp->if_capabilities |= IFCAP_CSUM_TCPv4;

	mii->mii_ifp = ifp;
	mii->mii_readreg = gmac_mii_readreg;
	mii->mii_writereg = gmac_mii_writereg;
	mii->mii_statchg = gmac_mii_statchg;

	ifmedia_init(&mii->mii_media, 0, gmac_mediachange, gmac_mediastatus);
	mii_attach(self, mii, 0xffffffff, MII_PHY_ANY, MII_OFFSET_ANY, flags);

	/* Choose a default media. */
	if (LIST_FIRST(&mii->mii_phys) == NULL) {
		ifmedia_add(&mii->mii_media, IFM_ETHER|IFM_NONE, 0, NULL);
		ifmedia_set(&mii->mii_media, IFM_ETHER|IFM_NONE);
	} else
		ifmedia_set(&mii->mii_media, IFM_ETHER|IFM_AUTO);

	if_attach(ifp);
	ether_ifattach(ifp, laddr);

	powerhook_establish(gmac_powerhook, sc);
}

u_int
gmac_read_reg(sc, reg)
	struct gmac_softc *sc;
	int reg;
{
	return in32rb(sc->sc_reg + reg);
}

void
gmac_write_reg(sc, reg, val)
	struct gmac_softc *sc;
	int reg;
	u_int val;
{
	out32rb(sc->sc_reg + reg, val);
}

void
gmac_start_txdma(sc)
	struct gmac_softc *sc;
{
	u_int x;

	x = gmac_read_reg(sc, GMAC_TXDMACONFIG);
	x |= 1;
	gmac_write_reg(sc, GMAC_TXDMACONFIG, x);
	x = gmac_read_reg(sc, GMAC_TXMACCONFIG);
	x |= 1;
	gmac_write_reg(sc, GMAC_TXMACCONFIG, x);
}

void
gmac_start_rxdma(sc)
	struct gmac_softc *sc;
{
	u_int x;

	x = gmac_read_reg(sc, GMAC_RXDMACONFIG);
	x |= 1;
	gmac_write_reg(sc, GMAC_RXDMACONFIG, x);
	x = gmac_read_reg(sc, GMAC_RXMACCONFIG);
	x |= 1;
	gmac_write_reg(sc, GMAC_RXMACCONFIG, x);
}

void
gmac_stop_txdma(sc)
	struct gmac_softc *sc;
{
	u_int x;

	x = gmac_read_reg(sc, GMAC_TXDMACONFIG);
	x &= ~1;
	gmac_write_reg(sc, GMAC_TXDMACONFIG, x);
	x = gmac_read_reg(sc, GMAC_TXMACCONFIG);
	x &= ~1;
	gmac_write_reg(sc, GMAC_TXMACCONFIG, x);
}

void
gmac_stop_rxdma(sc)
	struct gmac_softc *sc;
{
	u_int x;

	x = gmac_read_reg(sc, GMAC_RXDMACONFIG);
	x &= ~1;
	gmac_write_reg(sc, GMAC_RXDMACONFIG, x);
	x = gmac_read_reg(sc, GMAC_RXMACCONFIG);
	x &= ~1;
	gmac_write_reg(sc, GMAC_RXMACCONFIG, x);
}

int
gmac_intr(v)
	void *v;
{
	struct gmac_softc *sc = v;
	u_int status;

	status = gmac_read_reg(sc, GMAC_STATUS) & 0xff;
	if (status == 0)
		return 0;

	if (status & GMAC_INT_RXDONE)
		gmac_rint(sc);

	if (status & GMAC_INT_TXEMPTY)
		gmac_tint(sc);

	return 1;
}

void
gmac_tint(sc)
	struct gmac_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	int done, i, n;

	ifp->if_flags &= ~IFF_OACTIVE;
	ifp->if_timer = 0;

	i = sc->sc_txfree;
	done = gmac_read_reg(sc, GMAC_TXDMACOMPLETE);
	n = ((done - i - 1) & (NTXBUF - 1)) + 1;

	while (n > 0) {
		if (sc->sc_txbuf[i]) {
			m_freem(sc->sc_txbuf[i]);
			sc->sc_txbuf[i] = NULL;
		}
		i = (i + 1) & (NTXBUF - 1);
		n--;
	}
	sc->sc_txfree = i;

	gmac_start(ifp);
}

void
gmac_rint(sc)
	struct gmac_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	volatile struct gmac_dma *dp;
	struct mbuf *m, *n;
	int i, j, len;
	u_int cmd;

	for (i = sc->sc_rxlast;; i = (i + 1) & (NRXBUF - 1)) {
		dp = &sc->sc_rxlist[i];
		cmd = le32toh(dp->cmd);
		if (cmd & GMAC_OWN)
			break;

		if (le32toh(dp->cmd_hi) & GMAC_BADCRC) {
			ifp->if_ierrors++;
			goto next;
		}

		len = (cmd & GMAC_LEN_MASK) >> 16;
#ifdef DIAGNOSTIC
		if (len > MCLBYTES) {
			printf("%s: too big packet (%d)\n", ifp->if_xname, len);
			ifp->if_ierrors++;
			goto next;
		}
#endif

		m = sc->sc_rxbuf[i];
		m->m_len = m->m_pkthdr.len = len;

		n = gmac_get(sc);
		if (n == NULL) {
			ifp->if_ierrors++;
			goto next;
		}
		sc->sc_rxbuf[i] = n;

		/*
		 * Use hardware TCP checksum on large packets, since it's not
		 * very efficient for short packets.
		 */
		if (ifp->if_capenable & IFCAP_CSUM_TCPv4) {
			if (len > 100)
				gmac_rxhwsum(m, cmd & GMAC_CKSUM_MASK);
		}

#if NBPFILTER > 0
		/*
		 * Check if there's a BPF listener on this interface.
		 * If so, hand off the raw packet to BPF.
		 */
		if (ifp->if_bpf)
			bpf_mtap(ifp->if_bpf, m);
#endif
		(*ifp->if_input)(ifp, m);
		ifp->if_ipackets++;

#ifdef PMAP_UNMAP_POOLPAGE
		dp->address = htole32(PMAP_UNMAP_POOLPAGE(mtod(n, vaddr_t)));
#else
		dp->address = htole32(vtophys(mtod(n, vaddr_t)));
#endif

next:
		dp->cmd_hi = 0;
		asm volatile ("sync" ::: "memory");
		dp->cmd = htole32(GMAC_OWN);
	}
	sc->sc_rxlast = i;

	/* XXX Make sure free buffers have GMAC_OWN. */
	i++;
	for (j = 1; j < NRXBUF; j++) {
		i &= (NRXBUF - 1);
		dp = &sc->sc_rxlist[i++];
		if (dp->cmd != htole32(GMAC_OWN))
			dp->cmd = htole32(GMAC_OWN);
	}
}

void
gmac_rxhwsum(m, hwsum)
	struct mbuf *m;
	u_int hwsum;
{
	caddr_t packet = mtod(m, caddr_t);
	struct ether_header *ea = (void *)packet;
	struct ip *ip = (void *)(packet + 14);
	u_int32_t crc = *(u_int32_t *)(packet + m->m_pkthdr.len - 4);
	u_int16_t *w;
	u_int sum;
	int hlen;

#ifdef GMAC_HWCKSUM_TEST
	sum = in_cksum(m, m->m_pkthdr.len);
	if (sum != hwsum)
		printf("DIFFER: sum 0x%x (hwsum 0x%x)\n", sum, hwsum);
#endif

	/* XXX byte order */
	if (ea->ether_type != ETHERTYPE_IP)
		return;
	if (ip->ip_v != 4)
		return;
	if (ip->ip_p != IPPROTO_TCP)
		return;
	if (ip->ip_off & (IP_OFFMASK | IP_MF))
		return;

	/* Add phdr sum to hwsum. */
	hlen = ip->ip_hl << 2;
	hwsum ^= 0xffff;
	hwsum += in_cksum_phdr(ip->ip_src.s_addr, ip->ip_dst.s_addr,
	    IPPROTO_TCP + ip->ip_len - hlen);

	/* Link header and FCS checksum */
	w = (void *)packet;
	sum = 0;
	sum += w[0]; sum += w[1]; sum += w[2];
	sum += w[3]; sum += w[4]; sum += w[5];
	sum += w[6];

	if (m->m_pkthdr.len & 1)
		crc = crc << 8 | crc >> 24;	/* rotl crc,8 */
	sum += (crc >> 16) + (crc & 0xffff);
	sum = (sum >> 16) + (sum & 0xffff);
	sum += sum >> 16;
	sum = ~sum & 0xffff;

	/* Add them to hwsum; hwsum will be 0 if TCP checksum is correct. */
	hwsum += sum;
	hwsum = (hwsum >> 16) + (hwsum & 0xffff);
	hwsum += hwsum >> 16;
	hwsum = ~hwsum & 0xffff;

	m->m_pkthdr.csum_flags |= M_CSUM_TCPv4;
	if (hwsum) {
		m->m_pkthdr.csum_flags |= M_CSUM_TCP_UDP_BAD;
		printf("%s: bad checksum\n", m->m_pkthdr.rcvif->if_xname);
	}
}

struct mbuf *
gmac_get(sc)
	struct gmac_softc *sc;
{
	struct mbuf *m;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
		return NULL;
	m->m_pkthdr.rcvif = &sc->sc_if;
	m->m_flags |= M_HASFCS;

	MCLGET(m, M_DONTWAIT);
	if ((m->m_flags & M_EXT) == 0) {
		m_free(m);
		return NULL;
	}

	return m;
}

void
gmac_start(ifp)
	struct ifnet *ifp;
{
	struct gmac_softc *sc = ifp->if_softc;
	struct mbuf *m;
	int i, n, done, len, cmd;
	volatile struct gmac_dma *dp;

	if ((ifp->if_flags & (IFF_RUNNING | IFF_OACTIVE)) != IFF_RUNNING)
		return;

	/* Calculate free segments. */
	i = sc->sc_txnext;
	done = sc->sc_txfree;
	n = ((done - i - 1) & (NTXBUF - 1)) + 1;

start:
	i = sc->sc_txnext;

	IFQ_DEQUEUE(&ifp->if_snd, m);
	if (m == 0)
		return;

#if NBPFILTER > 0
	/*
	 * If BPF is listening on this interface, let it see the packet
	 * before we commit it to the wire.
	 */
	if (ifp->if_bpf)
		bpf_mtap(ifp->if_bpf, m);
#endif

#ifdef DIAGNOSTIC
	if ((m->m_flags & M_PKTHDR) == 0) {
		printf("%s: mbuf w/o M_PKTHDR\n", ifp->if_xname);
		goto start;
	}
#endif

	cmd = GMAC_SOF;
	if (m->m_pkthdr.csum_flags & M_CSUM_TCPv4) {
		struct ip *ip = (void *)(mtod(m, caddr_t) + ETHER_HDR_LEN);
		int tcphdr = ETHER_HDR_LEN + (ip->ip_hl << 2);
		int sum = tcphdr + m->m_pkthdr.csum_data;

		cmd |= GMAC_CKSUM | GMAC_CKSUM_PARAM(sum, tcphdr);

		/* XXX range check (sum, tcphdr) */
		/* XXX fragments? */
	}

	sc->sc_txbuf[i] = m;
	for (; m; m = m->m_next) {
		vaddr_t va = mtod(m, vaddr_t);
		int resid = m->m_len;

		while (resid > 0) {
			len = min(resid, NBPG - (va & PGOFSET));
			dp = &sc->sc_txlist[i];
			dp->cmd_hi = 0;
			dp->address_hi = 0;
			dp->cmd = htole32(len | cmd);
			cmd = 0;
			dp->address = htole32(vtophys(va));

			i = (i + 1) & (NTXBUF - 1);
			n--;

			va += len;
			resid -= len;
			if (resid == 0 && m->m_next == NULL)
				goto done;

			if (n == 0) {
				ifp->if_flags |= IFF_OACTIVE;

				/* XXX */
				printf("%s: dropped\n", ifp->if_xname);
				m_freem(sc->sc_txbuf[sc->sc_txnext]);
				return;
			}
		}
	}
done:
	if (cmd)
		goto start;

	dp->cmd |= htole32(GMAC_EOF);
	asm volatile ("sync" ::: "memory");

	/* 5 seconds to watch for failing to transmit */
	ifp->if_timer = 5;
	ifp->if_opackets++;		/* # of pkts */

	gmac_write_reg(sc, GMAC_TXDMAKICK, i);
	sc->sc_txnext = i;

	/* ĤꥻȤʤʤä */
	if (n < 8) {
		ifp->if_flags |= IFF_OACTIVE;
		return;
	}

	goto start;
}

void
gmac_reset(sc)
	struct gmac_softc *sc;
{
	int i, s;

	s = splnet();

	gmac_stop_txdma(sc);
	gmac_stop_rxdma(sc);

	gmac_write_reg(sc, GMAC_SOFTWARERESET, 3);
	for (i = 1000; i > 0; i--) {
		if ((gmac_read_reg(sc, GMAC_SOFTWARERESET) & 3) == 0)
			goto reset_done;
		delay(1000);
	}
	printf("%s: reset timeout\n", sc->sc_dev.dv_xname);

reset_done:
	sc->sc_txnext = 0;
	sc->sc_txfree = 0;
	for (i = 0; i < NTXBUF; i++)
		if (sc->sc_txbuf[i]) {
			m_freem(sc->sc_txbuf[i]);
			sc->sc_txbuf[i] = NULL;
		}
	sc->sc_rxlast = 0;
	for (i = 0; i < NRXBUF; i++)
		sc->sc_rxlist[i].cmd = htole32(GMAC_OWN);
	asm volatile ("sync");

	gmac_write_reg(sc, GMAC_TXDMADESCBASEHI, 0);
	gmac_write_reg(sc, GMAC_TXDMADESCBASELO,
		       vtophys((vaddr_t)sc->sc_txlist));
	gmac_write_reg(sc, GMAC_RXDMADESCBASEHI, 0);
	gmac_write_reg(sc, GMAC_RXDMADESCBASELO,
		       vtophys((vaddr_t)sc->sc_rxlist));
	gmac_write_reg(sc, GMAC_RXDMAKICK, NRXBUF);

	splx(s);
}

void
gmac_stop(sc)
	struct gmac_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	int s;

	s = splnet();

	callout_stop(&sc->sc_tick_ch);
	mii_down(&sc->sc_mii);

	gmac_stop_txdma(sc);
	gmac_stop_rxdma(sc);

	gmac_write_reg(sc, GMAC_INTMASK, 0xffffffff);

	ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
	ifp->if_timer = 0;

	splx(s);
}

#if   NTXBUF == 32
#define TXRINGSIZE GMAC_RINGSIZE_32
#elif NTXBUF == 64
#define TXRINGSIZE GMAC_RINGSIZE_64
#elif NTXBUF == 128
#define TXRINGSIZE GMAC_RINGSIZE_128
#elif NTXBUF == 256
#define TXRINGSIZE GMAC_RINGSIZE_256
#elif NTXBUF == 512
#define TXRINGSIZE GMAC_RINGSIZE_512
#elif NTXBUF == 1024
#define TXRINGSIZE GMAC_RINGSIZE_1024
#elif NTXBUF == 2048
#define TXRINGSIZE GMAC_RINGSIZE_2048
#elif NTXBUF == 4096
#define TXRINGSIZE GMAC_RINGSIZE_4096
#elif NTXBUF == 8192
#define TXRINGSIZE GMAC_RINGSIZE_8192
#else
#error NTXBUF
#endif

#if   NRXBUF == 32
#define RXRINGSIZE GMAC_RINGSIZE_32
#elif NRXBUF == 64
#define RXRINGSIZE GMAC_RINGSIZE_64
#elif NRXBUF == 128
#define RXRINGSIZE GMAC_RINGSIZE_128
#elif NRXBUF == 256
#define RXRINGSIZE GMAC_RINGSIZE_256
#elif NRXBUF == 512
#define RXRINGSIZE GMAC_RINGSIZE_512
#elif NRXBUF == 1024
#define RXRINGSIZE GMAC_RINGSIZE_1024
#elif NRXBUF == 2048
#define RXRINGSIZE GMAC_RINGSIZE_2048
#elif NRXBUF == 4096
#define RXRINGSIZE GMAC_RINGSIZE_4096
#elif NRXBUF == 8192
#define RXRINGSIZE GMAC_RINGSIZE_8192
#else
#error NRXBUF
#endif



void
gmac_init_mac(sc)
	struct gmac_softc *sc;
{
	int i;
	u_int tb, txmac, macctr, xif;
	char *laddr = sc->sc_laddr;

	asm ("mftb %0" : "=r"(tb));
	gmac_write_reg(sc, GMAC_RANDOMSEED, tb);

	/* init-mii */
	gmac_write_reg(sc, GMAC_DATAPATHMODE, DATAPATHMODE_GMII);

	gmac_write_reg(sc, GMAC_TXDMACONFIG, 0xffc00 | TXRINGSIZE);
	gmac_write_reg(sc, GMAC_RXDMACONFIG, RXRINGSIZE);
	gmac_write_reg(sc, GMAC_MACPAUSE, 0x1bf0);
	gmac_write_reg(sc, GMAC_INTERPACKETGAP0, 0);
	gmac_write_reg(sc, GMAC_INTERPACKETGAP1, 8);
	gmac_write_reg(sc, GMAC_INTERPACKETGAP2, 4);
	gmac_write_reg(sc, GMAC_MINFRAMESIZE, ETHER_MIN_LEN);
	gmac_write_reg(sc, GMAC_MAXFRAMESIZE, ETHER_MAX_LEN);
	gmac_write_reg(sc, GMAC_PASIZE, 7);
	gmac_write_reg(sc, GMAC_JAMSIZE, 4);
	gmac_write_reg(sc, GMAC_ATTEMPTLIMIT,0x10);
	gmac_write_reg(sc, GMAC_MACCNTLTYPE, 0x8808);

	gmac_write_reg(sc, GMAC_MACADDRESS0, (laddr[4] << 8) | laddr[5]);
	gmac_write_reg(sc, GMAC_MACADDRESS1, (laddr[2] << 8) | laddr[3]);
	gmac_write_reg(sc, GMAC_MACADDRESS2, (laddr[0] << 8) | laddr[1]);
	gmac_write_reg(sc, GMAC_MACADDRESS3, 0);
	gmac_write_reg(sc, GMAC_MACADDRESS4, 0);
	gmac_write_reg(sc, GMAC_MACADDRESS5, 0);
	gmac_write_reg(sc, GMAC_MACADDRESS6, 1);
	gmac_write_reg(sc, GMAC_MACADDRESS7, 0xc200);
	gmac_write_reg(sc, GMAC_MACADDRESS8, 0x0180);
	gmac_write_reg(sc, GMAC_MACADDRFILT0, 0);
	gmac_write_reg(sc, GMAC_MACADDRFILT1, 0);
	gmac_write_reg(sc, GMAC_MACADDRFILT2, 0);
	gmac_write_reg(sc, GMAC_MACADDRFILT2_1MASK, 0);
	gmac_write_reg(sc, GMAC_MACADDRFILT0MASK, 0);

	for (i = 0; i < 0x6c; i += 4)
		gmac_write_reg(sc, GMAC_HASHTABLE0 + i, 0);

	gmac_write_reg(sc, GMAC_SLOTTIME, 0x40);

	if (IFM_OPTIONS(sc->sc_mii.mii_media_active) & IFM_FDX) {
		txmac = TXMACCONFIG_IGNORE_CARR | TXMACCONFIG_IGNORE_COLL;
		xif = XIFCONFIG_MII_OE;
	} else {
		txmac = 0;
		xif = XIFCONFIG_MII_OE | XIFCONFIG_DISABLE_ECHO;
	}
	if (IFM_SUBTYPE(sc->sc_mii.mii_media_active) == IFM_1000_T)
		xif |=  XIFCONFIG_GMII_MODE;

	/* XXX Assume phy 0 */
	if (gmac_mii_readreg(&sc->sc_dev, 0, MII_ANLPAR) & ANLPAR_FC)
		macctr = MACCTRLCONFIG_SEND_PAUSE | MACCTRLCONFIG_RECV_PAUSE;
	else
		macctr = MACCTRLCONFIG_RECV_PAUSE;

	gmac_write_reg(sc, GMAC_TXMACCONFIG, txmac);
	gmac_write_reg(sc, GMAC_XIFCONFIG, xif);
	gmac_write_reg(sc, GMAC_MACCTRLCONFIG, macctr);
}

void
gmac_setladrf(sc)
	struct gmac_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	struct ether_multi *enm;
	struct ether_multistep step;
	struct ethercom *ec = &sc->sc_ethercom;
	u_int32_t crc;
	u_int32_t hash[16];
	u_int v;
	int i;

	/* Clear hash table */
	bzero(hash, sizeof hash);

	/* Get current RX configuration */
	v = gmac_read_reg(sc, GMAC_RXMACCONFIG);

	if ((ifp->if_flags & IFF_PROMISC) != 0) {
		/* Turn on promiscuous mode; turn off the hash filter */
		v |= GMAC_RXMAC_PR;
		v &= ~GMAC_RXMAC_HEN;
		ifp->if_flags |= IFF_ALLMULTI;
		goto chipit;
	}

	/* Turn off promiscuous mode; turn on the hash filter */
	v &= ~GMAC_RXMAC_PR;
	v |= GMAC_RXMAC_HEN;

	/*
	 * Set up multicast address filter by passing all multicast addresses
	 * through a crc generator, and then using the high order 8 bits as an
	 * index into the 256 bit logical address filter.  The high order bit
	 * selects the word, while the rest of the bits select the bit within
	 * the word.
	 */

	ETHER_FIRST_MULTI(step, ec, enm);
	while (enm != NULL) {
		if (bcmp(enm->enm_addrlo, enm->enm_addrhi, 6)) {
			/*
			 * We must listen to a range of multicast addresses.
			 * For now, just accept all multicasts, rather than
			 * trying to set only those filter bits needed to match
			 * the range.  (At this time, the only use of address
			 * ranges is for IP multicast routing, for which the
			 * range is big enough to require all bits set.)
			 */
			for (i = 0; i < 16; i++)
				hash[i] = 0xffff;
			ifp->if_flags |= IFF_ALLMULTI;
			goto chipit;
		}

		crc = ether_crc32_le(enm->enm_addrlo, ETHER_ADDR_LEN);

		/* Just want the 8 most significant bits. */
		crc >>= 24;

		/* Set the corresponding bit in the filter. */
		hash[crc >> 4] |= 0x8000 >> (crc & 0xf);

		ETHER_NEXT_MULTI(step, enm);
	}

	ifp->if_flags &= ~IFF_ALLMULTI;

chipit:
	/* Now load the hash table into the chip */
	for (i = 0; i < 16; i++)
		gmac_write_reg(sc, GMAC_HASHTABLE0 + i * 4, hash[i]);

	gmac_write_reg(sc, GMAC_RXMACCONFIG, v);
}

int
gmac_init(ifp)
	struct ifnet *ifp;
{
	struct gmac_softc *sc = ifp->if_softc;

	gmac_stop_txdma(sc);
	gmac_stop_rxdma(sc);

	/* Clear collision counter. */
	gmac_write_reg(sc, GMAC_NORMAL_COLLISION, 0);
	gmac_write_reg(sc, GMAC_FIRST_COLLISION, 0);
	gmac_write_reg(sc, GMAC_LATE_COLLISION, 0);

	gmac_init_mac(sc);
	gmac_setladrf(sc);
	gmac_mediachange(ifp);

	gmac_start_txdma(sc);
	gmac_start_rxdma(sc);

	gmac_write_reg(sc, GMAC_INTMASK, ~(GMAC_INT_TXEMPTY | GMAC_INT_RXDONE));

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;
	ifp->if_timer = 0;

	callout_reset(&sc->sc_tick_ch, 1, gmac_tick, sc);

	gmac_start(ifp);

	return 0;
}

int
gmac_ioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	caddr_t data;
{
	struct gmac_softc *sc = ifp->if_softc;
	struct ifreq *ifr = (struct ifreq *)data;
	int s, error = 0;

	s = splnet();

	switch (cmd) {

	case SIOCSIFADDR:
		ether_ioctl(ifp, cmd, data);
		break;

	case SIOCSIFFLAGS:
		if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) == IFF_RUNNING) {
			/*
			 * If interface is marked down and it is running, then
			 * stop it.
			 */
			gmac_stop(sc);
			ifp->if_flags &= ~IFF_RUNNING;
		} else if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) == IFF_UP) {
			/*
			 * If interface is marked up and it is stopped, then
			 * start it.
			 */
			gmac_init(ifp);
		} else {
			/*
			 * Reset the interface to pick up changes in any other
			 * flags that affect hardware registers.
			 */
			gmac_reset(sc);
			gmac_init(ifp);
		}
#ifdef GMAC_DEBUG
		if (ifp->if_flags & IFF_DEBUG)
			sc->sc_flags |= GMAC_DEBUGFLAG;
#endif
		break;

	case SIOCADDMULTI:
	case SIOCDELMULTI:
		error = ether_ioctl(ifp, cmd, data);
		if (error == ENETRESET) {
			/*
			 * Multicast list has changed; set the hardware filter
			 * accordingly.
			 */
			gmac_init(ifp);
			/* gmac_setladrf(sc); */
			error = 0;
		}
		break;

	case SIOCGIFMEDIA:
	case SIOCSIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->sc_mii.mii_media, cmd);
		break;

	default:
		error = EINVAL;
	}

	splx(s);
	return error;
}

void
gmac_watchdog(ifp)
	struct ifnet *ifp;
{
	struct gmac_softc *sc = ifp->if_softc;

	printf("%s: device timeout\n", ifp->if_xname);
	ifp->if_oerrors++;

	gmac_reset(sc);
	gmac_init(ifp);
}

int
gmac_mediachange(ifp)
	struct ifnet *ifp;
{
	struct gmac_softc *sc = ifp->if_softc;

	return mii_mediachg(&sc->sc_mii);
}

void
gmac_mediastatus(ifp, ifmr)
	struct ifnet *ifp;
	struct ifmediareq *ifmr;
{
	struct gmac_softc *sc = ifp->if_softc;

	mii_pollstat(&sc->sc_mii);

	ifmr->ifm_status = sc->sc_mii.mii_media_status;
	ifmr->ifm_active = sc->sc_mii.mii_media_active;
}

int
gmac_mii_readreg(dev, phy, reg)
	struct device *dev;
	int phy, reg;
{
	struct gmac_softc *sc = (void *)dev;
	int i;

	gmac_write_reg(sc, GMAC_MIFFRAMEOUTPUT,
		0x60020000 | (phy << 23) | (reg << 18));

	for (i = 1000; i >= 0; i -= 10) {
		if (gmac_read_reg(sc, GMAC_MIFFRAMEOUTPUT) & 0x10000)
			break;
		delay(10);
	}
	if (i < 0) {
		printf("%s: gmac_mii_readreg: timeout\n", sc->sc_dev.dv_xname);
		return 0;
	}

	return gmac_read_reg(sc, GMAC_MIFFRAMEOUTPUT) & 0xffff;
}

void
gmac_mii_writereg(dev, phy, reg, val)
	struct device *dev;
	int phy, reg, val;
{
	struct gmac_softc *sc = (void *)dev;
	int i;

	gmac_write_reg(sc, GMAC_MIFFRAMEOUTPUT,
		0x50020000 | (phy << 23) | (reg << 18) | (val & 0xffff));

	for (i = 1000; i >= 0; i -= 10) {
		if (gmac_read_reg(sc, GMAC_MIFFRAMEOUTPUT) & 0x10000)
			break;
		delay(10);
	}
	if (i < 0)
		printf("%s: gmac_mii_writereg: timeout\n", sc->sc_dev.dv_xname);
}

void
gmac_mii_statchg(dev)
	struct device *dev;
{
	struct gmac_softc *sc = (void *)dev;
	u_int txmac, macctr, xif;

	gmac_stop_txdma(sc);
	gmac_stop_rxdma(sc);

	if (IFM_OPTIONS(sc->sc_mii.mii_media_active) & IFM_FDX) {
		txmac = TXMACCONFIG_IGNORE_CARR | TXMACCONFIG_IGNORE_COLL;
		xif = XIFCONFIG_MII_OE;
	} else {
		txmac = 0;
		xif = XIFCONFIG_MII_OE | XIFCONFIG_DISABLE_ECHO;
	}
	if (IFM_SUBTYPE(sc->sc_mii.mii_media_active) == IFM_1000_T)
		xif |=  XIFCONFIG_GMII_MODE;

	if (gmac_mii_readreg(dev, 0, MII_ANLPAR) & ANLPAR_FC) /* XXX phy 0 */
		macctr = MACCTRLCONFIG_SEND_PAUSE | MACCTRLCONFIG_RECV_PAUSE;
	else
		macctr = MACCTRLCONFIG_RECV_PAUSE;

	gmac_write_reg(sc, GMAC_TXMACCONFIG, txmac);
	gmac_write_reg(sc, GMAC_XIFCONFIG, xif);
	gmac_write_reg(sc, GMAC_MACCTRLCONFIG, macctr);

	gmac_start_txdma(sc);
	gmac_start_rxdma(sc);
}

void
gmac_tick(v)
	void *v;
{
	struct gmac_softc *sc = v;
	struct ifnet *ifp = &sc->sc_if;
	int coll, s;

	s = splnet();
	mii_tick(&sc->sc_mii);

	coll = gmac_read_reg(sc, GMAC_NORMAL_COLLISION);
	gmac_write_reg(sc, GMAC_NORMAL_COLLISION, 0);
	ifp->if_collisions += coll;
	splx(s);

	callout_reset(&sc->sc_tick_ch, hz, gmac_tick, sc);
}

void
gmac_powerhook(why, v)
	int why;
	void *v;
{
	struct gmac_softc *sc = v;
	struct ifnet *ifp = &sc->sc_if;
	volatile int *gmac_gbclock_en = (void *)0xf8000020;
	int s;

	s = splnet();
	switch (why) {

	case PWR_SUSPEND:
		break;

	case PWR_RESUME:
		*gmac_gbclock_en |= 0x02;
		asm volatile ("eieio");
		gmac_reset(sc);
		gmac_init(ifp);
		break;
	}
	splx(s);
}
