/*	$Id: if_mc.c,v 1.17 2003/03/07 19:11:51 tsubai Exp $	*/

/*-
 * Copyright (c) 1997 David Huang <khym@bga.com>
 * All rights reserved.
 *
 * Portions of this code are based on code by Denton Gentry <denny1@home.com>
 * and Yanagisawa Takeshi <yanagisw@aa.ap.titech.ac.jp>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * Bus attachment and DMA routines for the mc driver (Centris/Quadra
 * 660av and Quadra 840av onboard ethernet, based on the AMD Am79C940
 * MACE ethernet chip). Also uses the PSC (Peripheral Subsystem
 * Controller) for DMA to and from the MACE.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/systm.h>

#include <net/if.h>
#include <net/if_ether.h>
#include <net/if_media.h>

#include <uvm/uvm_extern.h>

#include <dev/ofw/openfirm.h>

#include <machine/autoconf.h>
#include <machine/bus.h>
#include <machine/pio.h>

#include <macppc/dev/if_mcvar.h>
#include <macppc/dev/macereg.h>

#define MC_MTU (ETHER_MAX_LEN + 4)	/* ETHER_MAX_LEN + status word */

hide int	mc_match __P((struct device *, struct cfdata *, void *));
hide void	mc_attach __P((struct device *, struct device *, void *));
hide void	mc_init __P((struct mc_softc *));
hide void	mc_putpacket __P((struct mc_softc *, u_int));
hide int	mc_dmaintr __P((void *));
hide void	mc_reset_rxdma __P((struct mc_softc *));
hide void	mc_reset_txdma __P((struct mc_softc *));
hide int	mc_mediachange __P((struct ifnet *));
hide void	mc_mediastatus __P((struct ifnet *, struct ifmediareq *));

int mc_supmedia[] = {
	IFM_ETHER | IFM_10_T,
	IFM_ETHER | IFM_10_5,
};

#define N_SUPMEDIA (sizeof mc_supmedia / sizeof mc_supmedia[0])

struct cfattach mc_ca = {
	"mc", {}, sizeof(struct mc_softc), mc_match, mc_attach
};

hide int
mc_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "mace") != 0)
		return 0;

	/* requires 6 regs */
	if (ca->ca_nreg != 24)
		return 0;

	/* requires 3 intrs */
	if (ca->ca_nintr != 12)
		return 0;

	return 1;
}

hide void
mc_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct confargs *ca = aux;
	struct mc_softc *sc = (struct mc_softc *)self;
	u_int8_t myaddr[ETHER_ADDR_LEN];
	char compat[32];
	u_int *reg;
	int i;

#ifdef DIAGNOSTIC
	if ((vaddr_t)sc->sc_txdmacmd & 0x0f) {
		printf(": bad dbdma alignment\n");
		return;
	}
#endif
	sc->sc_node = ca->ca_node;

	reg  = ca->ca_reg;
	reg[0] += ca->ca_baseaddr;
	reg[2] += ca->ca_baseaddr;
	reg[4] += ca->ca_baseaddr;

	sc->sc_reg = mapiodev(reg[0], reg[1]);
	sc->sc_txdma = mapiodev(reg[2], reg[3]);
	sc->sc_rxdma = mapiodev(reg[4], reg[5]);

	sc->sc_tail = 0;

	printf(" irq %d,%d", ca->ca_intr[0], ca->ca_intr[2]);

	if (OF_getprop(sc->sc_node, "local-mac-address", myaddr, 6) != 6) {
		printf(": failed to get MAC address.\n");
		return;
	}

	/* Allocate memory for transmit buffer. */
	sc->sc_txbuf = malloc(NBPG, M_DEVBUF, M_WAITOK | M_ZERO);
	sc->sc_txbuf_phys = vtophys((vaddr_t)sc->sc_txbuf);
	if ((vaddr_t)sc->sc_txbuf & PGOFSET)
		printf(": txbuf is not page-aligned\n");

	/* Allocate mbufs for receive buffer. */
	for (i = 0; i < MC_RXDMABUFS; i++)
		sc->sc_rxbuf[i] = mace_get(sc);

	sc->sc_bus_init = mc_init;
	sc->sc_putpacket = mc_putpacket;

	/* Disable receive DMA. */
	dbdma_reset(sc->sc_rxdma);

	/* Disable transmit DMA. */
	dbdma_reset(sc->sc_txdma);

	/* Install interrupt handlers. */
	intr_establish(ca->ca_intr[0], IST_LEVEL, IPL_NET, mcintr, sc);
	intr_establish(ca->ca_intr[2], IST_LEVEL, IPL_NET, mc_dmaintr, sc);

	/* Initialize ifmedia structures. */
	ifmedia_init(&sc->sc_media, 0, mc_mediachange, mc_mediastatus);
	ifmedia_add(&sc->sc_media, IFM_ETHER|IFM_MANUAL, 0, NULL);
	ifmedia_set(&sc->sc_media, IFM_ETHER|IFM_MANUAL);

	sc->sc_biucc = XMTSP_64;
	sc->sc_fifocc = XMTFW_16 | RCVFW_64 | XMTFWU | RCVFWU |
	    XMTBRST | RCVBRST;
	bzero(compat, sizeof compat);
	OF_getprop(OF_finddevice("/"), "compatible", compat, sizeof compat);
	if (strcmp(compat, "AAPL,ShinerESB") == 0)
		sc->sc_plscc = PORTSEL_AUI | ENPLSIO;	/* -- for ANS */
	else
		sc->sc_plscc = PORTSEL_GPSI | ENPLSIO;
	/* mcsetup returns 1 if something fails */
	if (mcsetup(sc, myaddr)) {
		printf(": mcsetup returns non zero\n");
		return;
	}
}

/* Bus-specific initialization */
hide void
mc_init(sc)
	struct mc_softc *sc;
{
	mc_reset_rxdma(sc);
	mc_reset_txdma(sc);
}

hide void
mc_putpacket(sc, len)
	struct mc_softc *sc;
	u_int len;
{
	struct dbdma_desc *cmd = sc->sc_txdmacmd;

	dbdma_cmd(cmd, DBDMA_OUT_LAST, sc->sc_txbuf_phys, len, 0);
	dbdma_start(sc->sc_txdma, sc->sc_txdmacmd);
}

/*
 * Interrupt handler for the MACE DMA completion interrupts
 */
int
mc_dmaintr(arg)
	void *arg;
{
	struct mc_softc *sc = arg;
	struct ifnet *ifp = &sc->sc_if;
	struct dbdma_desc *cmd;
	struct mbuf *m;
	struct mc_rxframe *rxf;
	int len, status, resid, n, i;

	/* We've received some packets from the MACE */

	/* Loop through, processing each of the packets */
	i = sc->sc_tail;
	for (n = 0; n < MC_RXDMABUFS; n++, i++) {
		if (i == MC_RXDMABUFS)
			i = 0;

		cmd = &sc->sc_rxdmacmd[i];
		status = dbdma_ld16(&cmd->dd_status);
		resid = dbdma_ld16(&cmd->dd_resid);

		/*if ((status & DBDMA_ACTIVE) == 0)*/
		if ((status & 0x40) == 0)
			continue;

		len = MC_MTU - resid;
		len -= 4;	/* status word */

		cmd->dd_command = htole16(DBDMA_STOP);
		asm volatile ("sync" ::: "memory");

		m = sc->sc_rxbuf[i];
		m->m_pkthdr.len = m->m_len = len;
		rxf = (void *)(mtod(m, caddr_t) + len);

		m = mace_get(sc);
		if (m == NULL) {
			ifp->if_ierrors++;
			goto next;
		}

		mc_rint(sc, sc->sc_rxbuf[i], rxf);
		sc->sc_rxbuf[i] = m;
#ifdef PMAP_UNMAP_POOLPAGE
		cmd->dd_address =
		    htole32(PMAP_UNMAP_POOLPAGE(mtod(m, vaddr_t)));
#else
		cmd->dd_address = htole32(vtophys(mtod(m, vaddr_t)));
#endif

next:
		cmd->dd_status = 0;
		cmd->dd_resid = 0;
		cmd->dd_command = htole16(DBDMA_IN_LAST | DBDMA_INTR);
		sc->sc_tail = i + 1;
	}

	dbdma_continue(sc->sc_rxdma);

	return 1;
}

hide void
mc_reset_rxdma(sc)
	struct mc_softc *sc;
{
	struct dbdma_desc *cmd = sc->sc_rxdmacmd;
	struct dbdma_reg *dmareg = sc->sc_rxdma;
	int i;
	u_int8_t maccc;

	/* Disable receiver, reset the DMA channels */
	maccc = NIC_GET(sc, MACE_MACCC);
	NIC_PUT(sc, MACE_MACCC, maccc & ~ENRCV);

	dbdma_stop(dmareg);
	dbdma_reset(dmareg);

	for (i = 0; i < MC_RXDMABUFS; i++) {
		dbdma_cmd(cmd++, DBDMA_IN_LAST | DBDMA_INTR,
		    vtophys(mtod(sc->sc_rxbuf[i], vaddr_t)), MC_MTU, 0);
	}

	dbdma_cmd(cmd++, DBDMA_NOP | DBDMA_BR, 0, 0,
	    vtophys((vaddr_t)sc->sc_rxdmacmd));

	dbdma_start(dmareg, sc->sc_rxdmacmd);

	sc->sc_tail = 0;

	/* Reenable receiver, reenable DMA */
	NIC_PUT(sc, MACE_MACCC, maccc);
}

hide void
mc_reset_txdma(sc)
	struct mc_softc *sc;
{
	struct dbdma_desc *cmd = sc->sc_txdmacmd;
	struct dbdma_reg *dmareg = sc->sc_txdma;
	u_int8_t maccc;

	/* disable transmitter */
	maccc = NIC_GET(sc, MACE_MACCC);
	NIC_PUT(sc, MACE_MACCC, maccc & ~ENXMT);

	dbdma_stop(dmareg);
	dbdma_reset(dmareg);

	dbdma_cmd(cmd++, DBDMA_OUT_LAST, sc->sc_txbuf_phys, 0, 0);
	dbdma_cmd(cmd, DBDMA_STOP, 0, 0, 0);

	/* restore old value */
	NIC_PUT(sc, MACE_MACCC, maccc);
}

int
mc_mediachange(ifp)
	struct ifnet *ifp;
{
#if 0
	struct mc_softc *sc = ifp->if_softc;
	struct ifmedia *ifm = &sc->sc_media;

	if (IFM_TYPE(ifm->ifm_media) != IFM_ETHER)
		return EINVAL;

	switch (IFM_SUBTYPE(ifm->ifm_media)) {

	case IFM_10_T:
		sc->sc_plscc = PORTSEL_GPSI | ENPLSIO;
		return 0;

	case IFM_10_5:
		sc->sc_plscc = PORTSEL_AUI | ENPLSIO;
		return 0;
	}
#endif

	return EINVAL;
}

void
mc_mediastatus(ifp, ifmr)
	struct ifnet *ifp;
	struct ifmediareq *ifmr;
{
}
