/*	$Id: lm87.c,v 1.7 2003/07/01 13:14:21 tsubai Exp $	*/

/*-
 * Copyright (c) 2003 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Data sheet:
 * http://www.national.com/ds/LM/LM87.pdf
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <dev/ofw/openfirm.h>
#include <dev/sysmon/sysmonvar.h>

#include <machine/autoconf.h>
#include <machine/uni-n.h>

enum {
	PCITEMP = 0,
	CPUFAN,
	PCIFAN,
	VCC,
	P5V,
	P12V,
	NSENSORS
};

struct xlm_softc {
	struct device sc_dev;
	void *sc_i2c;
	struct sysmon_envsys sc_sysmon;
	struct envsys_tre_data sc_data[NSENSORS];
	struct envsys_basic_info sc_info[NSENSORS];
};

int xlm_match(struct device *, struct cfdata *, void *);
void xlm_attach(struct device *, struct device *, void *);
int xlm_read(struct xlm_softc *, int);
void xlm_write(struct xlm_softc *, int, int);
static int sysmon_attach(struct xlm_softc *);
int xlm_gtredata(struct sysmon_envsys *, struct envsys_tre_data *);
int xlm_streinfo(struct sysmon_envsys *, struct envsys_basic_info *);

struct cfattach xlm_ca = {
	"xlm", {}, sizeof(struct xlm_softc), xlm_match, xlm_attach
};

int
xlm_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	/* XXX for now */
	if (strcmp(ca->ca_name, "pwm") != 0)
		return 0;
	if (OF_finddevice("/uni-n/i2c/lm87") == -1)
		return 0;

	return 1;
}

void
xlm_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct xlm_softc *sc = (void *)self;
	int ui2c, node, addr;

	ui2c = OF_finddevice("ui2c");
	if (ui2c == -1) {
		printf(": cannot find ui2c node\n");
		return;
	}

	node = getnodebyname(ui2c, "lm87");
	if (node == 0) {
		printf(": cannot find lm87 node\n");
		return;
	}
	if (OF_getprop(node, "reg", &addr, 4) != 4) {
		printf(": cannot get my address\n");
		return;
	}

	sc->sc_i2c = ki2c_open(ui2c, addr, I2C_STDMODE);

	if (sc->sc_i2c == NULL) {
		printf(": cannot initialize i2c\n");
		return;
	}
	printf("\n");

	/* Set fan divisor to 4. */
	xlm_write(sc, 0x47, 0xa0);

	sysmon_attach(sc);
}

static struct envsys_range lm87_ranges[] = {
	{ 1, 0, ENVSYS_STEMP },
	{ 1, 0, ENVSYS_SFANRPM },
	{ 1, 0, ENVSYS_SVOLTS_AC },
	{ 1, 0, ENVSYS_SVOLTS_DC },
	{ 1, 0, ENVSYS_SOHMS },
	{ 1, 0, ENVSYS_SWATTS },
	{ 1, 0, ENVSYS_SAMPS },
};

int
sysmon_attach(sc)
	struct xlm_softc *sc;
{
	int i;

	/* Initialize sensors. */
	strcpy(sc->sc_info[PCITEMP].desc, "pci-temp");
	strcpy(sc->sc_info[CPUFAN].desc, "cpufan");
	strcpy(sc->sc_info[PCIFAN].desc, "pcifan");
	strcpy(sc->sc_info[VCC].desc, "+3.3V");
	strcpy(sc->sc_info[P5V].desc, "+5V");
	strcpy(sc->sc_info[P12V].desc, "+12V");

	sc->sc_data[PCITEMP].units = ENVSYS_STEMP;
	sc->sc_data[CPUFAN].units = ENVSYS_SFANRPM;
	sc->sc_data[PCIFAN].units = ENVSYS_SFANRPM;
	sc->sc_data[VCC].units = ENVSYS_SVOLTS_DC;
	sc->sc_data[P5V].units = ENVSYS_SVOLTS_DC;
	sc->sc_data[P12V].units = ENVSYS_SVOLTS_DC;

	for (i = 0; i < NSENSORS; i++) {
		sc->sc_data[i].sensor = i;
		sc->sc_data[i].validflags = ENVSYS_FVALID | ENVSYS_FCURVALID;
		sc->sc_data[i].warnflags = ENVSYS_WARN_OK;
		sc->sc_info[i].sensor = i;
		sc->sc_info[i].units = sc->sc_data[i].units;
		sc->sc_info[i].validflags = ENVSYS_FVALID;
	}

	sc->sc_sysmon.sme_nsensors = NSENSORS;
	sc->sc_sysmon.sme_envsys_version = 1000;
	sc->sc_sysmon.sme_ranges = lm87_ranges;
	sc->sc_sysmon.sme_sensor_data = sc->sc_data;
	sc->sc_sysmon.sme_sensor_info = sc->sc_info;
	sc->sc_sysmon.sme_gtredata = xlm_gtredata;
	sc->sc_sysmon.sme_streinfo = xlm_streinfo;
	sc->sc_sysmon.sme_cookie = sc;

	return sysmon_envsys_register(&sc->sc_sysmon);
}

int
xlm_read(sc, reg)
	struct xlm_softc *sc;
	int reg;
{
	u_char data[4];

	data[0] = reg;
	ki2c_write(sc->sc_i2c, data, 1);
	bzero(data, sizeof data);
	ki2c_read(sc->sc_i2c, data, 1);

	return data[0];
}

void
xlm_write(sc, reg, val)
	struct xlm_softc *sc;
	int reg, val;
{
	u_char data[4];

	data[0] = reg;
	data[1] = val;
	ki2c_write(sc->sc_i2c, data, 2);
}

int
xlm_gtredata(sme, tred)
	struct sysmon_envsys *sme;
	struct envsys_tre_data *tred;
{
	struct xlm_softc *sc = sme->sme_cookie;

	if (tred->sensor >= NSENSORS)
		return EINVAL;
	if ((sc->sc_info[tred->sensor].validflags & ENVSYS_FVALID) == 0)
		return 0;

	/* Refresh sensors. */
	switch (tred->sensor) {

	case PCITEMP:	/* Internal Temperature */
		sc->sc_data[PCITEMP].cur.data_s =
		    xlm_read(sc, 0x27) * 1000000 + 273150000;
		break;

	case CPUFAN:	/* FAN2/AIN2 */
		/*  2600rpm  6500rpm ޤѲ */
		sc->sc_data[CPUFAN].cur.data_us =
		    1350000 / 4 / xlm_read(sc, 0x29);
		break;

	case PCIFAN:	/* FAN1/AIN1 */
		sc->sc_data[PCIFAN].cur.data_us =
		    1350000 / 4 / xlm_read(sc, 0x28);
		break;

	case VCC:	/* +Vcc */
		sc->sc_data[VCC].cur.data_s =
		    xlm_read(sc, 0x22) * 3300000 / 192;
		break;

	case P5V:	/* +5V */
		sc->sc_data[P5V].cur.data_s =
		    xlm_read(sc, 0x23) * 5000000 / 192;
		break;

	case P12V:	/* +12V */
		sc->sc_data[P12V].cur.data_s =
		    xlm_read(sc, 0x24) * 12000000 / 192;
		break;
	}
	*tred = sc->sc_data[tred->sensor];
	return 0;
}

int
xlm_streinfo(sme, binfo)
	struct sysmon_envsys *sme;
	struct envsys_basic_info *binfo;
{
	printf("streinfo sensor = %d\n", binfo->sensor);

	binfo->validflags = 0;
	return 0;
}
