/*	$Id: nvram.c,v 1.12 2003/03/28 11:31:10 tsubai Exp $	*/

/*-
 * Copyright (c) 2002 Tsubai Masanari.
 * Copyright (c) 1998 Internet Research Institute, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by
 *	Internet Research Institute, Inc.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/malloc.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/pio.h>

#define NVRAM_NONE  0
#define NVRAM_IOMEM 1
#define NVRAM_PORT  2
#define NVRAM_FLASH 3

void nvram_attach __P((struct device *, struct device *, void *));
int nvram_match __P((struct device *, struct cfdata *, void *));
int nvram_read_brightness(void);

struct nvram_softc {
	struct device sc_dev;
	int sc_nvsize;
	int sc_nvtype;
	char *sc_nvport;
	char *sc_nvdata;
};

struct cfattach nvram_obio_ca = {
	"nvram_obio", {}, sizeof(struct nvram_softc), nvram_match, nvram_attach
};
struct cfattach nvram_mainbus_ca = {
	"nvram_mainbus", {}, sizeof(struct nvram_softc),
	nvram_match, nvram_attach
};

extern struct cfdriver nvram_cd;

dev_type_read(nvramread);
dev_type_write(nvramwrite);
dev_type_mmap(nvrammmap);

const struct cdevsw nvram_cdevsw = {
	nullopen, nullclose, nvramread, nvramwrite, noioctl,
	nostop, notty, nopoll, nvrammmap, nokqfilter
};

int
nvram_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "nvram") != 0)
		return 0;

	if (ca->ca_nreg == 0)
		return 0;

	return 1;
}

void
nvram_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct nvram_softc *sc = (struct nvram_softc *)self;
	struct confargs *ca = aux;
	int *reg = ca->ca_reg;


	if (OF_compat(ca->ca_compat, "nvram,flash")) {
		sc->sc_nvtype = NVRAM_FLASH;
		sc->sc_nvdata = (void *)reg[0];
		sc->sc_nvsize = reg[1];
		printf(": addr %p, size 0x%x\n", sc->sc_nvdata, sc->sc_nvsize);
		return;
	}

	printf("\n");

	switch (ca->ca_nreg) {

	case 8:						/* untested */
		sc->sc_nvtype = NVRAM_IOMEM;
		sc->sc_nvsize = 0x2000;
		sc->sc_nvdata = mapiodev(ca->ca_baseaddr + reg[0], reg[1]);
		break;

	case 16:
		sc->sc_nvtype = NVRAM_PORT;
		sc->sc_nvsize = 0x2000;
		sc->sc_nvport = mapiodev(ca->ca_baseaddr + reg[0], reg[1]);
		sc->sc_nvdata = mapiodev(ca->ca_baseaddr + reg[2], reg[3]);
		break;

	case 0:
	default:
		sc->sc_nvtype = NVRAM_NONE;
		return;
	}
}

int
nvramread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct nvram_softc *sc;
	u_int off, cnt;
	int i;
	int error = 0;
	char *buf;

	sc = nvram_cd.cd_devs[0];

	off = uio->uio_offset;
	cnt = uio->uio_resid;

	if (off > sc->sc_nvsize)
		return EFAULT;

	if (off + cnt > sc->sc_nvsize)
		cnt = sc->sc_nvsize - off;

	buf = malloc(sc->sc_nvsize, M_DEVBUF, M_WAITOK);

	switch (sc->sc_nvtype) {

	case NVRAM_FLASH:
		bcopy(sc->sc_nvdata, buf, sc->sc_nvsize);
		break;

	case NVRAM_IOMEM:
		for (i = 0; i < sc->sc_nvsize; i++)
			buf[i] = sc->sc_nvdata[i * 16];
		break;

	case NVRAM_PORT:
		for (i = 0; i < sc->sc_nvsize; i += 32) {
			int j;

			out8(sc->sc_nvport, i / 32);
			for (j = 0; j < 32; j++) {
				buf[i + j] = sc->sc_nvdata[j * 16];
			}
		}
		break;

	default:
		error = EINVAL;
		goto out;
	}

	error = uiomove(buf + off, cnt, uio);

out:
	free(buf, M_DEVBUF);
	return error;
}

int
nvramwrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	return ENXIO;
}

paddr_t
nvrammmap(dev, off, prot)
        dev_t dev;
        off_t off;
	int prot;
{
	return -1;
}

/*
 * Below is NewWorld nvram format.
 */
struct nv_header {
	u_int8_t magic;
	u_int8_t sum;
	u_int16_t len;
	char name[12];
};

struct nv_nvram {
	struct nv_header nvh;
	u_int32_t checksum;
	u_int16_t reserved;
	u_int16_t serial;
};

struct nv_common {
	struct nv_header nvh;
	char data[0x7f0];		/* XXX variable length */
};

int
nvram_read_brightness()
{
	struct nvram_softc *sc = nvram_cd.cd_devs[0];	/* XXX */
	struct nv_header *nvh;
	struct nv_nvram *nvram;
	struct nv_common *common;
	int off, len, serial, p_nvram, p_common;
	u_char *p;

	if (sc == NULL || sc->sc_nvtype != NVRAM_FLASH)
		return -1;

	off = 0;
	serial = -1;
	p_nvram = p_common = -1;

	/* Find active block. */
	while (off < sc->sc_nvsize) {
		nvh = (void *)&sc->sc_nvdata[off];

		if (strcmp(nvh->name, "nvram") == 0) {
			nvram = (void *)nvh;
			if (nvram->serial > serial) {
				serial = nvram->serial;
				p_nvram = off;
			}
		}
		off += nvh->len << 4;
	}
	if (p_nvram == -1)
		return -1;

	/* Find "common" partition in active block. */
	off = p_nvram;
	while (off < p_nvram + 0x2000) {
		nvh = (void *)&sc->sc_nvdata[off];

		if (strcmp(nvh->name, "common") == 0)
			p_common = off;
		off += nvh->len << 4;
	}
	if (p_common == -1)
		return -1;

	/* Find "bklt" in "common" partition. */
	common = (void *)&sc->sc_nvdata[p_common];
	p = common->data;
	len = common->nvh.len << 4;
	while (len > 0) {
		if (strncmp(p, "bklt", 4) == 0)
			return p[6];
		p++;
		len--;
	}

	return -1;
}
