/*	$Id: uni-n.c,v 1.10 2003/07/04 09:54:37 tsubai Exp $	*/

/*-
 * Copyright (c) 2001 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/systm.h>

#include <dev/ofw/openfirm.h>
#include <uvm/uvm_extern.h>
#include <machine/uni-n.h>

/* Keywest I2C Register offsets */
#define MODE	0
#define CONTROL	1
#define STATUS	2
#define ISR	3
#define IER	4
#define ADDR	5
#define SUBADDR	6
#define DATA	7

/* MODE */
#define I2C_SPEED	0x03	/* Speed mask */
#define  I2C_100kHz	0x00
#define  I2C_50kHz	0x01
#define  I2C_25kHz	0x02
#define I2C_MODE	0x0c	/* Mode mask */
#define  I2C_DUMBMODE	0x00	/*  Dumb mode */
#define  I2C_STDMODE	0x04	/*  Standard mode */
#define  I2C_STDSUBMODE	0x08	/*  Standard mode + sub address */
#define  I2C_COMBMODE	0x0c	/*  Combined mode */
#define I2C_PORT	0xf0	/* Port mask */

/* CONTROL */
#define I2C_CT_AAK	0x01	/* Send AAK */
#define I2C_CT_ADDR	0x02	/* Send address(es) */
#define I2C_CT_STOP	0x04	/* Send STOP */
#define I2C_CT_START	0x08	/* Send START */

/* STATUS */
#define I2C_ST_BUSY	0x01	/* Busy */
#define I2C_ST_LASTAAK	0x02	/* Last AAK */
#define I2C_ST_LASTRW	0x04	/* Last R/W */
#define I2C_ST_SDA	0x08	/* SDA */
#define I2C_ST_SCL	0x10	/* SCL */

/* ISR/IER */
#define I2C_INT_DATA	0x01	/* Data byte sent/received */
#define I2C_INT_ADDR	0x02	/* Address sent */
#define I2C_INT_STOP	0x04	/* STOP condition sent */
#define I2C_INT_START	0x08	/* START condition sent */

/* I2C flags */
#define I2C_BUSY	0x01
#define I2C_READING	0x02
#define I2C_ERROR	0x04

struct ki2c_softc {
	u_char *sc_reg;
	int sc_regstep;

	int sc_flags;
	int sc_addr;	/* target addr */
	int sc_mode;	/* access mode */
	u_char *sc_data;
	int sc_resid;
};

inline u_int ki2c_readreg(struct ki2c_softc *, int);
inline void ki2c_writereg(struct ki2c_softc *, int, u_int);
int ki2c_intr(struct ki2c_softc *);
int ki2c_poll(struct ki2c_softc *, int);
int ki2c_start(struct ki2c_softc *, int, void *, int);

struct ki2c_softc *
ki2c_open(node, addr, mode)
	int node, addr, mode;
{
	struct ki2c_softc *sc;
	u_int rate, reg, regstep;

	if (OF_getprop(node, "AAPL,i2c-rate", &rate, 4) != 4)
		return NULL;
	if (OF_getprop(node, "AAPL,address", &reg, 4) != 4)
		return NULL;
	if (OF_getprop(node, "AAPL,address-step", &regstep, 4) != 4)
		return NULL;

	sc = malloc(sizeof *sc, M_DEVBUF, M_NOWAIT);
	if (sc == NULL)
		return NULL;

	sc->sc_reg = (void *)reg;
	sc->sc_regstep = regstep;

	ki2c_writereg(sc, STATUS, 0);
	ki2c_writereg(sc, ISR, 0xff);
	ki2c_writereg(sc, IER, 0);

	KASSERT((addr & 1) == 0);

	sc->sc_addr = addr & 0xff;
	sc->sc_mode = mode;
	sc->sc_mode |= (addr >> 8) << 4;	/* port number */
	if (mode == 0)
		sc->sc_mode = I2C_STDMODE|I2C_100kHz;	/* XXX */
	ki2c_writereg(sc, MODE, sc->sc_mode);

	return sc;
}

u_int
ki2c_readreg(sc, reg)
	struct ki2c_softc *sc;
	int reg;
{
	u_char *addr = sc->sc_reg + sc->sc_regstep * reg;

	return *addr;
}

void
ki2c_writereg(sc, reg, val)
	struct ki2c_softc *sc;
	int reg;
	u_int val;
{
	u_char *addr = sc->sc_reg + sc->sc_regstep * reg;

	*addr = val;
	asm volatile ("eieio");
	delay(10);
}

int
ki2c_intr(sc)
	struct ki2c_softc *sc;
{
	u_int isr, x;

	isr = ki2c_readreg(sc, ISR);

	if (isr & I2C_INT_ADDR) {
		if ((ki2c_readreg(sc, STATUS) & I2C_ST_LASTAAK) == 0) {
			/* No slave responded. */
			sc->sc_flags |= I2C_ERROR;
			goto out;
		}

		if (sc->sc_flags & I2C_READING) {
			if (sc->sc_resid > 1) {
				x = ki2c_readreg(sc, CONTROL);
				x |= I2C_CT_AAK;
				ki2c_writereg(sc, CONTROL, x);
			}
		} else {
			ki2c_writereg(sc, DATA, *sc->sc_data++);
			sc->sc_resid--;
		}
	}

	if (isr & I2C_INT_DATA) {
		if (sc->sc_flags & I2C_READING) {
			*sc->sc_data++ = ki2c_readreg(sc, DATA);
			sc->sc_resid--;

			if (sc->sc_resid == 0) {	/* Completed */
				ki2c_writereg(sc, CONTROL, 0);
				goto out;
			}
		} else {
			if ((ki2c_readreg(sc, STATUS) & I2C_ST_LASTAAK) == 0) {
				/* No slave responded. */
				sc->sc_flags |= I2C_ERROR;
				goto out;
			}

			if (sc->sc_resid == 0) {
				x = ki2c_readreg(sc, CONTROL) | I2C_CT_STOP;
				ki2c_writereg(sc, CONTROL, x);
			} else {
				ki2c_writereg(sc, DATA, *sc->sc_data++);
				sc->sc_resid--;
			}
		}
	}

out:
	if (isr & I2C_INT_STOP) {
		ki2c_writereg(sc, CONTROL, 0);
		sc->sc_flags &= ~I2C_BUSY;
	}

	ki2c_writereg(sc, ISR, isr);

	return 1;
}

int
ki2c_poll(sc, timo)
	struct ki2c_softc *sc;
	int timo;
{
	while (sc->sc_flags & I2C_BUSY) {
		if (ki2c_readreg(sc, ISR))
			ki2c_intr(sc);
		timo -= 100;
		if (timo < 0) {
			printf("i2c_poll: timeout\n");
			return -1;
		}
		delay(100);
	}
	return 0;
}

int
ki2c_start(sc, subaddr, data, len)
	struct ki2c_softc *sc;
	int subaddr, len;
	void *data;
{
	int addr = sc->sc_addr;
	int rw = (sc->sc_flags & I2C_READING) ? 1 : 0;
	int timo, x;

	/* Set port number, access mode, and speed. */
	ki2c_writereg(sc, MODE, sc->sc_mode);

	sc->sc_data = data;
	sc->sc_resid = len;
	sc->sc_flags |= I2C_BUSY;

	timo = 1000 + len * 200;

	/* XXX TAS3001 sometimes takes 50ms to finish writing registers. */
	/* if (addr == 0x68) */
		timo += 100000;

	ki2c_writereg(sc, ADDR, addr | rw);
	ki2c_writereg(sc, SUBADDR, subaddr);

	x = ki2c_readreg(sc, CONTROL) | I2C_CT_ADDR;
	ki2c_writereg(sc, CONTROL, x);

	if (ki2c_poll(sc, timo))
		return -1;
	if (sc->sc_flags & I2C_ERROR) {
		printf("I2C_ERROR\n");
		return -1;
	}
	return 0;
}

int
ki2c_read(sc, data, len)
	struct ki2c_softc *sc;
	int len;
	void *data;
{
	return ki2c_readsub(sc, 0, data, len);
}

int
ki2c_write(sc, data, len)
	struct ki2c_softc *sc;
	int len;
	const void *data;
{
	return ki2c_writesub(sc, 0, data, len);
}

int
ki2c_readsub(sc, subaddr, data, len)
	struct ki2c_softc *sc;
	int subaddr, len;
	void *data;
{
	sc->sc_flags = I2C_READING;
	return ki2c_start(sc, subaddr, data, len);
}

int
ki2c_writesub(sc, subaddr, data, len)
	struct ki2c_softc *sc;
	int subaddr, len;
	const void *data;
{
	sc->sc_flags = 0;
	return ki2c_start(sc, subaddr, (void *)data, len);
}

/* Uni-N registers */
#define UNIN_VERS	0x000		/* Uni-North version */
#define UNIN_CLOCK_CTRL	0x020		/* PCI clock control */
#define   UNIN_CLOCK_PCI2  0x01
#define   UNIN_CLOCK_ENET  0x02
#define   UNIN_CLOCK_FW    0x04
#define   UNIN_CLOCK_ATA   0x10
#define UNIN_POWER	0x030		/* Power Management */
#define   UNIN_POWER_NORM  0x00
#define   UNIN_POWER_IDLE2 0x01
#define   UNIN_POWER_SLEEP 0x02
#define UNIN_ARBCTRL	0x040
#define UNIN_STATE	0x070		/* Hardware initial state */
#define   UNIN_STATE_SLEEP 0x01
#define   UNIN_STATE_RUN   0x02

#define UNIN_MMCR	0xf00
#define UNIN_MCMDR	0xf10
#define UNIN_MPMC1	0xf20
#define UNIN_MPMC2	0xf30
#define UNIN_MPMC3	0xf40
#define UNIN_MPMC4	0xf50

/*
 * CPU Υåؤ롣ΤȤ 745x ꡣ(XXX)
 */
int macppc_suspend(void);
void change_clock(int);
extern long ticks_per_intr;

vaddr_t uninorth_base;

#include <macppc/dev/pm_direct.h>
#include <machine/bat.h>
#include <machine/cpufunc.h>
#include <machine/hid.h>
#include <machine/pio.h>
#include <machine/pte.h>
#include <machine/spr.h>

#define mtsrin(sr, addr) asm volatile ("mtsrin %0,%1" :: "r"(sr), "r"(addr))
#define mfsrin(sr, addr) asm ("mfsrin %0,%1" : "=r"(sr) : "r"(addr))

void
change_clock(step)
	int step;
{
	int l3cr, msr, pir, hid1, msscr0, sdr1, hid0, sr[16], sprg[4];
	struct bat bat[8];
	u_int64_t tb;
	int suspend, i, tmp;

	if (step == -1)
		suspend = 1;
	else
		suspend = 0;

	if (suspend)
		dopowerhooks(PWR_SUSPEND);

	msr = mfmsr();

	/* Disable interrupts. */
	mtmsr(msr & ~PSL_EE);
	asm volatile ("isync; sync");

	if (suspend)
		pm_sleep();
	else
		pm_speed_step(step);
	out32(uninorth_base + UNIN_STATE, UNIN_STATE_SLEEP);
	out32(uninorth_base + UNIN_POWER, UNIN_POWER_IDLE2);

	pir = mfspr(PIR);
	msscr0 = mfspr(MSSCR0);
	hid0 = mfspr(HID0);
	hid1 = mfspr(HID1);
	l3cr = mfspr(L3CR);
	asm volatile ("mfsprg %0,0" : "=r"(sprg[0]));
	asm volatile ("mfsprg %0,1" : "=r"(sprg[1]));
	asm volatile ("mfsprg %0,2" : "=r"(sprg[2]));
	asm volatile ("mfsprg %0,3" : "=r"(sprg[3]));

	asm volatile ("mfsdr1 %0" : "=r"(sdr1));
	for (i = 0; i < 16; i++)
		mfsrin(sr[i], i << ADDR_SR_SHFT);
	bat[0].batu = mfdbatu(0);
	bat[0].batl = mfdbatl(0);
	bat[1].batu = mfdbatu(1);
	bat[1].batl = mfdbatl(1);
	/* XXX bat4-7 */

	/* XXX msssr0? ldstcr, ldstdb, ictrl Ȥ */

	/* Select SLEEP mode. */
	tmp = hid0;
	tmp &= ~(HID0_DOZE|HID0_NAP);
	tmp |= HID0_SLEEP;
	mtspr(HID0, tmp);

	tb = mftb();

	/* XXX å񤭽Ф */
	for (i = 0; i < 65536*2*4; i++)
		in32(i * 32);

	macppc_suspend();

	/*
	 * ɤäƤ֤롣
	 */

	mtdec(0x7fffffff);
	asm volatile ("mttbl %0" :: "r"(0));
	asm volatile ("mttbu %0" :: "r"(tb >> 32));
	asm volatile ("mttbl %0" :: "r"(tb & 0xffffffff));

	mtspr(PIR, pir);
	asm volatile ("mtsprg 0,%0" :: "r"(sprg[0]));
	/* XXX sprg1-3 ʤƤ⤤Ϥ */

	mtspr(MSSCR0, msscr0);

	/* Enable the L1 caches. */
	mpc7450_enable_L1cache();

	mtspr(HID0, hid0);
	mtspr(HID1, hid1);

	/* Enable the L2 cache. */
	mpc7450_enable_L2cache();

	/* Enable the L3 cache. */
	mpc7450_enable_L3cache(l3cr);

	asm volatile ("mtsdr1 %0" :: "r"(sdr1));
	for (i = 0; i < 16; i++)
		mtsrin(sr[i], i << ADDR_SR_SHFT);
	asm volatile ("mtdbatu 0,%0; mtibatu 0,%0" :: "r"(bat[0].batu));
	asm volatile ("mtdbatl 0,%0; mtibatl 0,%0" :: "r"(bat[0].batl));
	asm volatile ("mtdbatu 1,%0; mtibatu 1,%0" :: "r"(bat[1].batu));
	asm volatile ("mtdbatl 1,%0; mtibatl 1,%0" :: "r"(bat[1].batl));
	asm volatile ("mtdbatu 2,%0; mtibatu 2,%0" :: "r"(0));
	asm volatile ("mtdbatu 3,%0; mtibatu 3,%0" :: "r"(0));
	/* XXX bat4-7 */

	/* Enable address translation. */
	asm volatile ("isync");
	mtmsr(msr & ~PSL_EE);
	asm volatile ("sync; isync");

	out32(uninorth_base + UNIN_POWER, UNIN_POWER_NORM);
	delay(10);
	out32(uninorth_base + UNIN_STATE, UNIN_STATE_RUN);
	delay(10);

	mpc7450_print_hid("cpu0");	/* XXX cpu0 */

	if (suspend) {
		void openpic_init(void);
		void intr_calculatemasks(void);
		u_int fcr2;

		/* Clear Keylargo sleep bit. */
		fcr2 = in32rb(0x80000040);
		fcr2 &= ~0x100;
		out32rb(0x80000040, fcr2);
		delay(10);

		openpic_init();
		intr_calculatemasks();
	}

	mtdec(ticks_per_intr);

	/* Enable interrupts. */
	mtmsr(msr);

	if (suspend)
		dopowerhooks(PWR_RESUME);
}

void unin_init(void);

void
unin_init()
{
	int node;
	u_int reg[2], x, vers;

	node = OF_finddevice("/uni-n");
	if (node == -1)
		return;
	if (OF_getprop(node, "reg", reg, sizeof reg) < 8)
		return;

	uninorth_base = reg[0];
	vers = in32(uninorth_base + UNIN_VERS);
	/* printf("uni-n version 0x%x\n", vers); */

	/* Set QACK delay. */
	if (vers < 0x11) {
		x = in32(uninorth_base + UNIN_ARBCTRL);
		x &= ~0x0e1f8000;
		if (vers >= 0x03)
			x |= 0x180000;
		out32(uninorth_base + UNIN_ARBCTRL, x);
	}

	if ((vers >= 0x11 && vers <= 0x24) || vers == 0xc0) {
		x = in32(uninorth_base + 0x2160);
		x &= 0x00ffffff;
		out32(uninorth_base + 0x2160, x);
	}
}

void unin_enable_clock(int);

void
unin_enable_clock(node)
	int node;
{
	volatile int *clkctl = (void *)(uninorth_base + UNIN_CLOCK_CTRL);
	int child;
	char compat[32];

	/* Enable gmac and kauai-ata. */
	for (child = OF_child(node); child; child = OF_peer(child)) {
		bzero(compat, sizeof(compat));
		OF_getprop(child, "compatible", compat, sizeof compat);
		if (strcmp(compat, "gmac") == 0)
			*clkctl |= UNIN_CLOCK_ENET;
		if (strcmp(compat, "kauai-ata") == 0)
			*clkctl |= UNIN_CLOCK_ATA;
	}
}
