/*	$Id: clock.c,v 1.15 2002/04/29 11:58:44 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_timebase_freq.h"

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>
#include <dev/ofw/openfirm.h>

#include <machine/autoconf.h>
#include <machine/cpu.h>
#include <machine/cpufunc.h>

#include "adb.h"

/*
 * Initially we assume a processor with a bus frequency of 133 MHz.
 */
static u_long ticks_per_sec = 133*1000*1000/4;
static u_long ns_per_tick = 30;
long ticks_per_intr;
static int clockinitted;

#ifdef MULTIPROCESSOR
#define lasttb (curcpu()->ci_lasttb)
#define tickspending (curcpu()->ci_tickspending)
#else
static volatile u_long lasttb;
volatile int tickspending;
#endif

#ifdef TIMEBASE_FREQ
u_int timebase_freq = TIMEBASE_FREQ;
#else
u_int timebase_freq = 0;
#endif

#define SECDAY 86400
#define DIFF19041970 2082844800

#if NADB > 0
int adb_read_date_time __P((int *));
int adb_set_date_time __P((int));
#else
#define adb_read_date_time(t) (-1)
#define adb_set_date_time(t)
#endif

void decr_intr __P((struct clockframe *));
static void calibrate_clock(void);

void
inittodr(base)
	time_t base;
{
	time_t deltat;
	u_int rtc_time;

	/*
	 * If we can't read from RTC, use the fs time.
	 */
	if (adb_read_date_time(&rtc_time) < 0) {
		time.tv_sec = base;
		return;
	}
	clockinitted = 1;
	time.tv_sec = rtc_time - DIFF19041970 + rtc_offset * 60;

	deltat = time.tv_sec - base;
	if (deltat < 0)
		deltat = -deltat;
	if (deltat < 2 * SECDAY)
		return;

	printf("WARNING: clock %s %d days",
	    time.tv_sec < base ? "lost" : "gained", (int)(deltat / SECDAY));

	printf(" -- CHECK AND RESET THE DATE!\n");
}

/*
 * Similar to the above
 */
void
resettodr()
{
	u_int rtc_time;

	if (clockinitted) {
		rtc_time = time.tv_sec + DIFF19041970 - rtc_offset * 60;
		adb_set_date_time(rtc_time);
	}
}

void
decr_intr(frame)
	struct clockframe *frame;
{
	u_long tb;
	long tick;
	int nticks;
	int pri, msr;

	/*
	 * Check whether we are initialized.
	 */
	if (!ticks_per_intr)
		return;

	/*
	 * Based on the actual time delay since the last decrementer reload,
	 * we arrange for earlier interrupt next time.
	 */
	asm ("mfdec %0" : "=r"(tick));
	for (nticks = 0; tick < 0; nticks++)
		tick += ticks_per_intr;
	asm volatile ("mtdec %0" :: "r"(tick));

	uvmexp.intrs++;
	intrcnt[CNT_CLOCK]++;

	pri = splclock();
#ifdef NEWINTR
	if (pri >= IPL_CLOCK)
#else
	if (pri & (1 << SPL_CLOCK))
#endif
		tickspending += nticks;
	else {
		nticks += tickspending;
		tickspending = 0;

		/*
		 * lasttb is used during microtime. Set it to the virtual
		 * start of this tick interval.
		 */
		asm ("mftb %0" : "=r"(tb));
		lasttb = tb + tick - ticks_per_intr;

		/*
		 * Reenable interrupts
		 */
		asm volatile ("mfmsr %0; ori %0, %0, %1; mtmsr %0"
			      : "=r"(msr) : "K"(PSL_EE));

		/*
		 * Do standard timer interrupt stuff.
		 * Do softclock stuff only on the last iteration.
		 */
#ifdef NEWINTR
		frame->pri = IPL_HIGH;	/* XXX max(pri, IPL_SOFTCLOCK) */
#else
		frame->pri = pri | (1 << SIR_CLOCK);
#endif
		while (--nticks > 0)
			hardclock(frame);
		frame->pri = pri;
		hardclock(frame);
	}
	splx(pri);
}

void
cpu_initclocks()
{
}

void
calc_delay()
{
	int qhandle, phandle;
	char type[32];
	int msr, scratch;

	/*
	 * Get this info during autoconf?				XXX
	 */
	if (timebase_freq != 0) {
		ticks_per_sec = timebase_freq;
		goto found;
	}

	for (qhandle = OF_peer(0); qhandle; qhandle = phandle) {
		if (OF_getprop(qhandle, "device_type", type, sizeof type) > 0
		    && strcmp(type, "cpu") == 0
		    && OF_getprop(qhandle, "timebase-frequency",
			   &ticks_per_sec, sizeof ticks_per_sec) > 0) {
			goto found;
		}
		if ((phandle = OF_child(qhandle)))
			continue;
		while (qhandle) {
			if ((phandle = OF_peer(qhandle)))
				break;
			qhandle = OF_parent(qhandle);
		}
	}
	panic("no cpu node");

found:
	/*
	 * Should check for correct CPU here?		XXX
	 */
	asm volatile ("mfmsr %0; andi. %1,%0,%2; mtmsr %1"
		      : "=r"(msr), "=r"(scratch) : "K"((u_short)~PSL_EE));
	if (timebase_freq == 0)
		calibrate_clock();
	ns_per_tick = 1000000000 / ticks_per_sec;
	ticks_per_intr = (ticks_per_sec + hz / 2) / hz;
	asm volatile ("mftb %0" : "=r"(lasttb));
	asm volatile ("mtdec %0" :: "r"(ticks_per_intr));
	asm volatile ("mtmsr %0" :: "r"(msr));
}

/*
 * Fill in *tvp with current time with microsecond resolution.
 */
void
microtime(tvp)
	struct timeval *tvp;
{
	u_long tb;
	u_long ticks;
	int msr, scratch;

	asm volatile ("mfmsr %0; andi. %1,%0,%2; mtmsr %1"
		      : "=r"(msr), "=r"(scratch) : "K"((u_short)~PSL_EE));
	asm ("mftb %0" : "=r"(tb));
	ticks = (tb - lasttb) * ns_per_tick;
	*tvp = time;
	asm volatile ("mtmsr %0" :: "r"(msr));
	ticks /= 1000;
	tvp->tv_usec += ticks;
	while (tvp->tv_usec >= 1000000) {
		tvp->tv_usec -= 1000000;
		tvp->tv_sec++;
	}
}

/*
 * Wait for about n microseconds (at least!).
 */
void
delay(n)
	unsigned int n;
{
	u_quad_t tb;
	u_long tbh, tbl, scratch;

	tb = mftb();
	tb += (n * 1000 + ns_per_tick - 1) / ns_per_tick;
	tbh = tb >> 32;
	tbl = tb;
	asm volatile ("1: mftbu %0; cmplw %0,%1; blt 1b; bgt 2f;"
		      "mftb %0; cmplw %0,%2; blt 1b; 2:"
		      : "=r"(scratch) : "r"(tbh), "r"(tbl));
	asm volatile ("sync; isync");
}

/*
 * Nothing to do.
 */
void
setstatclockrate(arg)
	int arg;
{
	/* Do nothing */
}

#define KL_TIMER_LO 0x38
#define KL_TIMER_HI 0x3c
#define KL_TIMER_FREQ 18432000

#if 0
# define DPRINTF printf
#else
# define DPRINTF while (0) printf
#endif

void
calibrate_clock()
{
	u_int klbase, timerbase, timerhi, timerlo;
	u_int tbfreq, timerfreq;
	u_int tbh0, tbl0, tbh1, tbl1, klh0, kll0, klh1, kll1, tmp;
	u_int delta_tb, delta_kl;
	int keylargo, timer;
	int reg[5];

	if ((keylargo = getnodebyname(0, "mac-io")) == 0)
		return;
	if ((timer = getnodebyname(keylargo, "timer")) == 0)
		return;
	if (OF_getprop(keylargo, "assigned-addresses", reg, 20) < 12)
		return;
	klbase = reg[2];
	if (OF_getprop(timer, "reg", reg, 8) < 8)
		return;
	if (OF_getprop(timer, "clock-frequency", &timerfreq, 4) < 4)
		return;
	ns_per_tick = 1000000000 / ticks_per_sec;	/* XXX for delay() */

	timerbase = klbase + reg[0];
	timerhi = timerbase + KL_TIMER_HI;
	timerlo = timerbase + KL_TIMER_LO;

	/*
	 * Calculate timebase frequency by keylargo timer.
	 * ("timebase-frequecy" property is sometimes inaccurate)
	 */
retry:
	asm volatile ("mftbu %0" : "=r"(tbh0));
	asm volatile ("mftb %0" : "=r"(tbl0));
	asm volatile ("mftbu %0" : "=r"(tmp));
	if (tbh0 != tmp)
		goto retry;
	asm volatile ("lwbrx %0,0,%1" : "=r"(klh0) : "r"(timerhi));
	asm volatile ("lwbrx %0,0,%1" : "=r"(kll0) : "r"(timerlo));
	asm volatile ("lwbrx %0,0,%1" : "=r"(tmp) : "r"(timerhi));
	if (klh0 != tmp)
		goto retry;

	delay(500000);		/* 0.5 sec */

	asm volatile ("mftbu %0" : "=r"(tbh1));
	asm volatile ("mftb %0" : "=r"(tbl1));
	asm volatile ("mftbu %0" : "=r"(tmp));
	if (tbh1 != tmp)
		goto retry;
	asm volatile ("lwbrx %0,0,%1" : "=r"(klh1) : "r"(timerhi));
	asm volatile ("lwbrx %0,0,%1" : "=r"(kll1) : "r"(timerlo));
	asm volatile ("lwbrx %0,0,%1" : "=r"(tmp) : "r"(timerhi));
	if (klh1 != tmp)
		goto retry;

	delta_tb = ((u_int64_t)tbh1 << 32 | tbl1) -
		   ((u_int64_t)tbh0 << 32 | tbl0);
	delta_kl = ((u_int64_t)klh1 << 32 | kll1) -
		   ((u_int64_t)klh0 << 32 | kll0);

	DPRINTF("delta_tb = 0x%x\n", delta_tb);
	DPRINTF("delta_kl = 0x%x\n", delta_kl);

	if (delta_kl < 1000) {
		printf("keylargo counter is not running...\n");
		return;
	}
	tbfreq = (u_int64_t)delta_tb * timerfreq / delta_kl;
	DPRINTF("observed tbfreq = %d\n", tbfreq);

	/*
	 * If "timebase-frequency" supplied by OF is close enough to
	 * observed one, use the original (OF) value.
	 */
	if (abs(ticks_per_sec - tbfreq) < 1000)
		return;

	printf("timebase frequency: %d Hz\n", tbfreq);
	ticks_per_sec = tbfreq;
}
