/*	$Id: cpu.c,v 1.69 2003/09/30 10:26:20 tsubai Exp $	*/

/*-
 * Copyright (c) 2001, 2002 Tsubai Masanari.
 * Copyright (c) 1998, 1999, 2001 Internet Research Institute, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by
 *	Internet Research Institute, Inc.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "opt_multiprocessor.h"

#include <sys/param.h>
#include <sys/callout.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/reboot.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>
#include <dev/ofw/openfirm.h>
#include <powerpc/openpic.h>

#include <machine/autoconf.h>
#include <machine/bat.h>
#include <machine/cpufunc.h>
#include <machine/fpu.h>
#include <machine/hid.h>
#include <machine/pcb.h>
#include <machine/pio.h>
#include <machine/spr.h>
#include <machine/trap.h>

int cpumatch(struct device *, struct cfdata *, void *);
void cpuattach(struct device *, struct device *, void *);

static void ohare_init(void);
void config_l2cr(void);
void config_l3cr(void);
void cpu_powerctl(void *);
void cpu_powerctl_init(void);
int cpu_spinup(void *);
void cpu_hatch(void);

void cpu_spinup_trampoline(void);

struct cfattach cpu_ca = {
	"cpu", {}, sizeof(struct device), cpumatch, cpuattach
};

int ncpus;
static struct callout powerctl_ch;

#ifdef MULTIPROCESSOR
int cpuintr(void *);
struct cpu_info cpu_info[2];
#else
struct cpu_info cpu_info_store;
#endif

extern struct cfdriver cpu_cd;
extern int powersave;

#define HH_ARBCONF		0xf8000090
#define HH_INTR_SECONDARY	0xf80000c0
#define HH_INTR_PRIMARY		0xf3019000
#define GC_NMI_IRQ		20
#define GC_IPI_IRQ		30

int
cpumatch(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;
	int *reg = ca->ca_reg;
	int node;

	if (strcmp(ca->ca_name, cpu_cd.cd_name) != 0)
		return 0;

	node = OF_finddevice("/cpus");
	if (node != -1) {
		for (node = OF_child(node); node != 0; node = OF_peer(node)) {
			u_int32_t cpunum;
			int l;
			l = OF_getprop(node, "reg", &cpunum, sizeof(cpunum));
			if (l == 4 && reg[0] == cpunum)
				return 1;
		}
	}
	switch (reg[0]) {
	case 0:	/* primary CPU */
		return 1;
	case 1:	/* secondary CPU */
		if (OF_finddevice("/hammerhead") != -1)
			if (in32rb(HH_ARBCONF) & 0x02)
				return 1;
		break;
	}

	return 0;
}

void
cpuattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct confargs *ca = aux;
	int id = ca->ca_reg[0];
	u_int hid0, pvr, vers, msscr0;
	char model[80], hidbuf[128];

	ncpus++;
#ifdef MULTIPROCESSOR
	cpu_info[id].ci_cpuid = id;
	cpu_info[id].ci_intrdepth = -1;
#endif

	asm ("mfpvr %0" : "=r"(pvr));
	vers = pvr >> 16;
#ifdef PPC750FX_PLL0
	if (vers == MPC750)
		vers = PPC750FX;
#endif

	switch (id) {
	case 0:
		/* load my cpu_number to PIR */
		switch (vers) {
		case MPC604:
		case MPC604ev:
		case MPC7400:
		case MPC7410:
		case MPC7450:
		case MPC7455:
		case MPC7457:
			asm volatile ("mtspr PIR,%0" :: "r"(id));
		}
		identifycpu(model);
		printf(": %s, ID %d (primary)\n", model, cpu_number());
		break;
	case 1:
#ifdef MULTIPROCESSOR
		cpu_spinup(self);
#else
		printf(" not configured\n");
#endif
		return;
	default:
		printf(": more than 2 cpus?\n");
		panic("cpuattach");
	}

	asm ("mfspr %0,HID0" : "=r"(hid0));

	if (bootverbose) {
		switch (vers) {
		case PPC750FX:
			ppc750fx_print_hid(self->dv_xname);
			break;
		case MPC7450:
		case MPC7455:
		case MPC7457:
			mpc7450_print_hid(self->dv_xname);
			break;
		default:
			bitmask_snprintf(hid0, HID0_BITMASK, hidbuf,
			    sizeof hidbuf);
			printf("%s: HID0 %s\n", self->dv_xname, hidbuf);
		}
	}

	/*
	 * Configure power-saving mode.
	 */
	switch (vers) {
	case MPC603:
	case MPC603e:
	case MPC603ev:
	case MPC750:
	case PPC750FX:
	case MPC7400:
	case MPC7410:
		/* Select DOZE mode. */
		hid0 &= ~(HID0_DOZE | HID0_NAP | HID0_SLEEP);
		hid0 |= HID0_DOZE | HID0_DPM;
		powersave = 1;
		break;

	case MPC7450:
	case MPC7455:
	case MPC7457:
		/* XXX Does this mean DOZE mode? */
		hid0 &= ~(HID0_NAP | HID0_SLEEP);
		hid0 |= HID0_DPM;
		powersave = 1;
		break;

	default:
		/* No power-saving mode is available. */
		break;
	}

#ifdef NAPMODE
	switch (vers) {
	case MPC750:
	case MPC7400:
	case MPC7410:
		/* Select NAP mode. */
		hid0 &= ~(HID0_DOZE | HID0_NAP | HID0_SLEEP);
		hid0 |= HID0_NAP;
		break;

#if 0
	case MPC7450:
	case MPC7455:
	case MPC7457:
		/* Select NAP mode. */
		hid0 &= ~(HID0_NAP | HID0_SLEEP);
		hid0 |= HID0_NAP;
		break;
#endif
	}
#endif

	switch (vers) {
	case MPC750:
	case PPC750FX:
		hid0 &= ~HID0_DBP;		/* XXX correct? */
		hid0 &= ~HID0_ECLK;
		hid0 |= HID0_EMCP | HID0_BTIC | HID0_SGE | HID0_BHT;
		break;
	case MPC7400:
	case MPC7410:
		hid0 &= ~HID0_SPD;
		hid0 |= HID0_EMCP | HID0_BTIC | HID0_SGE | HID0_BHT;
		hid0 |= HID0_EIEC;
		break;
	case MPC7450:
	case MPC7455:
	case MPC7457:
		break;
	}

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));

	switch (vers) {
	case PPC750FX:
		ppc750fx_print_hid(self->dv_xname);
		break;
	case MPC7450:
	case MPC7455:
	case MPC7457:
		mpc7450_print_hid(self->dv_xname);
		break;
	default:
		bitmask_snprintf(hid0, HID0_BITMASK, hidbuf, sizeof hidbuf);
		printf("%s: HID0 %s\n", self->dv_xname, hidbuf);
	}

#define MSSCR0_SHDEN 0x80000000	/* Shared-state enable (MESI) */
	/*
	 * Enable MESI protocol on bandit based models.
	 */
	if (OF_finddevice("/bandit") != -1) {
		switch (vers) {
		case MPC7400:
		case MPC7410:

			asm ("mfspr %0,MSSCR0" : "=r"(msscr0));
			msscr0 |= MSSCR0_SHDEN;
			asm volatile ("mtspr MSSCR0,%0" :: "r"(msscr0));
		}
	}

	/*
	 * Configure L2/L3 cache.
	 */
	switch (vers) {
	case MPC750:
	case MPC7400:
	case MPC7410:
		printf("%s", self->dv_xname);
		config_l2cr();
		break;
	case PPC750FX:
		printf("%s", self->dv_xname);
		ppc750fx_config_l2cr();
		break;
	case MPC7450:
	case MPC7455:
	case MPC7457:
		printf("%s", self->dv_xname);
		config_l3cr();
		break;
	default:
		if (OF_finddevice("/bandit/ohare") != -1) {
			printf("%s", self->dv_xname);
			ohare_init();
		}
	}

	switch (vers) {
	case MPC7450:
	case MPC7455:
	case MPC7457:
		cpu_powerctl_init();
	}
}

struct cputab {
	u_int16_t version;
	char name[6];
};
static struct cputab models[] = {
	{ MPC601,   "601" },
	{ MPC602,   "602" },
	{ MPC603,   "603" },
	{ MPC603e,  "603e" },
	{ MPC603ev, "603ev" },
	{ MPC604,   "604" },
	{ MPC604ev, "604ev" },
	{ MPC620,   "620" },
	{ MPC750,   "750" },
	{ PPC750FX, "750FX" },
	{ MPC7400,  "7400" },
	{ MPC7410,  "7410" },
	{ MPC7450,  "7450" },
	{ MPC7455,  "7455" },
	{ MPC7457,  "7457" },
	{ 0,	    "" }
};

void
identifycpu(cpu_model)
	char *cpu_model;
{
	u_int pvr, vers, rev;
	struct cputab *cp = models;

	asm ("mfpvr %0" : "=r"(pvr));
	vers = pvr >> 16;
	rev = pvr & 0xffff;
#ifdef PPC750FX_PLL0
	if (vers == MPC750)
		vers = PPC750FX;
#endif

	while (cp->version) {
		if (cp->version == vers)
			break;
		cp++;
	}
	if (cp->version)
		strcpy(cpu_model, cp->name);
	else
		sprintf(cpu_model, "Version %x", vers);
	sprintf(cpu_model + strlen(cpu_model), " (Revision %x)", rev);
}

#define CACHE_REG 0xf8000000

void
ohare_init()
{
	u_int *cache_reg, x;

	/* enable L2 cache */
	cache_reg = mapiodev(CACHE_REG, NBPG);
	if (((cache_reg[2] >> 24) & 0x0f) >= 3) {
		x = cache_reg[4];
		if ((x & 0x10) == 0)
                	x |= 0x04000000;
		else
                	x |= 0x04000020;

		cache_reg[4] = x;
		printf(": ohare L2 cache enabled\n");
	}
}

#ifdef AUTONAP
void
cpu_powerctl(v)
	void *v;
{
	u_int hid0;
	static int lastintr, timo;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	if (uvmexp.intrs - lastintr - hz > 20) {
		/* HIGH-POWER mode */
		timo = 5;
		hid0 &= ~(HID0_NAP | HID0_SLEEP);
	} else {
		/* LOW-POWER (NAP) mode */

		/* ö H ˤʤääʤ褦ˤ */
		if (timo > 0)
			goto out;
		timo = 0;
		hid0 &= ~(HID0_NAP | HID0_SLEEP);
		hid0 |= HID0_NAP;
	}
	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
out:
	timo--;
	lastintr = uvmexp.intrs;
	callout_reset(&powerctl_ch, hz, cpu_powerctl, NULL);
}
#endif

void
cpu_powerctl_init()
{
	callout_init(&powerctl_ch);
#ifdef AUTONAP
	callout_reset(&powerctl_ch, 30*hz, cpu_powerctl, NULL);
#endif
}

#ifdef MULTIPROCESSOR

volatile struct cpu_hatch_data *cpu_hatch_data;
volatile int cpu_hatch_stack;

int
cpu_spinup(v)
	void *v;
{
	volatile struct cpu_hatch_data hatch_data, *h = &hatch_data;
	int i;
	struct pcb *pcb;
	struct pglist mlist;
	int error;
	int size = 0;
	char *cp;
	u_int pvr, vers;

	/*
	 * Allocate some contiguous pages for the idle PCB and stack
	 * from the lowest 256MB (because bat0 always maps it va == pa).
	 */
	size += USPACE;
	size += 8192;	/* INTSTK */
#if DDB | IPKDB
	size += 8192;	/* DDBSTK */
#endif
	size += 4096;	/* SPILLSTK */
	error = uvm_pglistalloc(size, 0x0, 0x10000000, 0, 0, &mlist, 1, 1);
	if (error) {
		printf(": unable to allocate idle stack\n");
		return -1;
	}

	cp = (void *)VM_PAGE_TO_PHYS(TAILQ_FIRST(&mlist));
	bzero(cp, size);

	pcb = (struct pcb *)cp;
	cp += USPACE;
	cpu_info[1].ci_idle_pcb = pcb;

	cpu_info[1].ci_intstk = cp + 8192;
	cp += 8192;
#if DDB | IPKDB
	cpu_info[1].ci_ddbstk = cp + 8192;
	cp += 8192;
#endif
	cpu_info[1].ci_spillstk = cp + 4096;
	cp += 4096;

	/*
	 * Initialize the idle stack pointer, reserving space for an
	 * (empty) trapframe (XXX is the trapframe really necessary?)
	 */
	pcb->pcb_sp = (paddr_t)pcb + USPACE - sizeof(struct trapframe);

	asm ("mfpvr %0" : "=r"(pvr));
	vers = pvr >> 16;

	cpu_hatch_data = h;
	h->dev = v;
	h->running = 0;
	h->pir = 1;
	cpu_hatch_stack = pcb->pcb_sp;
	cpu_info[1].ci_lasttb = cpu_info[0].ci_lasttb;

	/* Copy special registers. */
	asm volatile ("mfspr %0,HID0" : "=r"(h->hid0));
	/* XXX */
	if (vers == MPC7450 || vers == MPC7455 || vers == MPC7457) {
		asm volatile ("mfspr %0,HID1" : "=r"(h->hid1));
		asm volatile ("mfspr %0,MSSCR0" : "=r"(h->msscr0));
		asm volatile ("mfspr %0,L2CR" : "=r"(h->l2cr));
		asm volatile ("mfspr %0,L3CR" : "=r"(h->l3cr));
	}

	asm volatile ("mfsdr1 %0" : "=r"(h->sdr1));
	for (i = 0; i < 16; i++)
		asm ("mfsrin %0,%1" : "=r"(h->sr[i]) : "r"(i << ADDR_SR_SHFT));

	asm volatile ("sync; isync");

	if (openpic_base) {
		u_int kl_base = 0x80000000;	/* XXX */
		u_int gpio = kl_base + 0x5c;	/* XXX */
		/* cpu0  cpu1  cpu2  cpu3 */
		/* 0x5b, 0x5c, 0x67, 0x68 */
		u_quad_t tb;
		int node;
		u_int off;

		*(u_int *)EXC_RST =		/* ba cpu_spinup_trampoline */
		    0x48000002 | (u_int)cpu_spinup_trampoline;
		__syncicache((void *)EXC_RST, 0x100);

		h->running = -1;

		/* Start secondary cpu. */
		node = OF_finddevice("/cpus/@1");	/* XXX CPU 1 */
		if (node == -1) {
			printf(": no secondary cpu node?\n");
			return -1;
		}
		if (OF_getprop(node, "soft-reset", &off, 4) == 4)
			gpio = kl_base + off;

		out8(gpio, 4);
		out8(gpio, 5);

		/* Sync timebase. */
		tb = mftb();
		tb += 100000;	/* 3ms @ 33MHz */

		h->tbu = tb >> 32;
		h->tbl = tb & 0xffffffff;

		while (tb > mftb());
		asm volatile ("sync; isync");
		h->running = 0;

		delay(500000);
	} else {
		/* Start secondary cpu and stop timebase. */
		out32(0xf2800000, (int)cpu_spinup_trampoline);
		out32(HH_INTR_SECONDARY, ~0);
		out32(HH_INTR_SECONDARY, 0);

		/* sync timebase (XXX shouldn't be zero'ed) */
		asm volatile ("mttbl %0; mttbu %0; mttbl %0" :: "r"(0));

		/*
		 * wait for secondary spin up (1.5ms @ 604/200MHz)
		 * XXX we cannot use delay() here because timebase is not
		 * running.
		 */
		for (i = 0; i < 100000; i++)
			if (h->running)
				break;

		/* Start timebase counter. */
		out32(0xf2800000, 0x100);
		out32(HH_INTR_SECONDARY, ~0);
		out32(HH_INTR_SECONDARY, 0);
	}

	delay(100000);		/* wait for secondary printf */

	if (h->running == 0) {
		printf(": secondary cpu didn't start\n");
		return -1;
	}

	/* Register IPI. */
	if (!openpic_base)
		intr_establish(GC_IPI_IRQ, IST_LEVEL, IPL_HIGH, cpuintr, NULL);

	return 0;
}

volatile static int start_secondary_cpu;
extern long ticks_per_intr;

void
cpu_hatch()
{
	volatile struct cpu_hatch_data *h = cpu_hatch_data;
	u_int msr, pvr, vers;
	int i;
	char model[80];

	asm ("mfpvr %0" : "=r"(pvr));
	vers = pvr >> 16;

	/* Initialize timebase. */
	asm ("mttbl %0; mttbu %0; mttbl %0" :: "r"(0));

	if (vers == MPC7450 || vers == MPC7455 || vers == MPC7457)
		mpc7450_initialize();

	/* Set PIR (Processor Identification Register).  i.e. whoami */
	asm volatile ("mtspr PIR,%0" :: "r"(h->pir));
	asm volatile ("mtsprg 0,%0" :: "r"(&cpu_info[h->pir]));

	/* Initialize MMU. */
	asm ("mtibatu 0,%0" :: "r"(0));
	asm ("mtibatu 1,%0" :: "r"(0));
	asm ("mtibatu 2,%0" :: "r"(0));
	asm ("mtibatu 3,%0" :: "r"(0));
	asm ("mtdbatu 0,%0" :: "r"(0));
	asm ("mtdbatu 1,%0" :: "r"(0));
	asm ("mtdbatu 2,%0" :: "r"(0));
	asm ("mtdbatu 3,%0" :: "r"(0));

	asm ("mtspr HID0,%0" :: "r"(h->hid0));

	asm ("mtibatl 0,%0; mtibatu 0,%1;"
	     "mtdbatl 0,%0; mtdbatu 0,%1;"
		:: "r"(battable[0].batl), "r"(battable[0].batu));

	/* BAT1 used for secondary 256 MB segment (if exists) */
	asm volatile ("mtibatl 1,%0; mtibatu 1,%1;"
		      "mtdbatl 1,%0; mtdbatu 1,%1;"
		      :: "r"(battable[1].batl), "r"(battable[1].batu));

	/*
	 * Use extended BAT size on 745[57].
	 */
	switch (vers) {

	case MPC7455:
	case MPC7457:
		/* mpc7455_xbsen(); */	/* Already done. */
		asm volatile ("mtibatl 0,%0; mtibatu 0,%1;"
			      "mtdbatl 0,%0; mtdbatu 0,%1;"
		    :: "r"(battable[0].batl), "r"(battable[0].batu));
		asm volatile ("mtibatu 1,%0" :: "r"(0));
		asm volatile ("mtdbatu 1,%0" :: "r"(0));
		break;
	}

	for (i = 0; i < 16; i++)
		asm ("mtsrin %0,%1" :: "r"(h->sr[i]), "r"(i << ADDR_SR_SHFT));
	asm ("mtsdr1 %0" :: "r"(h->sdr1));

	asm volatile ("isync");

	/* Enable I/D address translations. */
	asm volatile ("mfmsr %0" : "=r"(msr));
	msr |= PSL_IR|PSL_DR|PSL_ME|PSL_RI;
	asm volatile ("mtmsr %0" :: "r"(msr));

	asm volatile ("sync; isync");

	if (openpic_base) {
		/* Sync timebase. */
		u_int tbu = h->tbu;
		u_int tbl = h->tbl;
		while (h->running == -1);
		asm volatile ("sync; isync");
		asm volatile ("mttbl %0" :: "r"(0));
		asm volatile ("mttbu %0" :: "r"(tbu));
		asm volatile ("mttbl %0" :: "r"(tbl));
	}

	h->running = 1;

	identifycpu(model);
	printf(": %s, ID %d\n", model, cpu_number());

	switch (vers) {

	case MPC7400:
	case MPC7410:
		printf("%s", h->dev->dv_xname);
		config_l2cr();
		break;
	case MPC7450:
	case MPC7455:
	case MPC7457:
		mpc7450_print_hid(h->dev->dv_xname);
		printf("%s", h->dev->dv_xname);
		config_l3cr();
		break;
	}

	while (start_secondary_cpu == 0);

	printf("secondary CPU started\n");
	asm volatile ("mtdec %0" :: "r"(ticks_per_intr));

	if (openpic_base)
		openpic_set_priority(cpu_number(), 0);
	else
		out32(HH_INTR_SECONDARY, ~0);	/* Reset interrupt. */

	curcpu()->ci_ipending = 0;
	curcpu()->ci_cpl = 0;
}

void
cpu_boot_secondary_processors()
{
	start_secondary_cpu = 1;
}

/*static*/ volatile int IPI[2];
static struct simplelock ipilock = SIMPLELOCK_INITIALIZER;

void
macppc_send_ipi(ci, mesg)
	volatile struct cpu_info *ci;
	int mesg;
{
	int cpu_id = ci->ci_cpuid;
	int s;

	/* printf("send_ipi(%d,%d)\n", cpu_id, mesg); */
	s = splhigh();
	simple_lock(&ipilock);
	IPI[cpu_id] |= mesg;
	simple_unlock(&ipilock);
	splx(s);

	if (openpic_base) {
		/* XXX */

		openpic_write(OPENPIC_IPI_VECTOR(1),
		    15 << OPENPIC_PRIORITY_SHIFT | 240);
		openpic_write(OPENPIC_IPI(cpu_number(), 1), 1 << cpu_id);
	} else {
		switch (cpu_id) {
		case 0:
			in32(HH_INTR_PRIMARY);
			break;
		case 1:
			out32(HH_INTR_SECONDARY, ~0);
			out32(HH_INTR_SECONDARY, 0);
			break;
		}
	}
}

int
cpuintr(v)
	void *v;
{
	int cpu_id = cpu_number();
	int ipi = IPI[cpu_id];
	int s, msr;

	/* printf("cpuintr{%d}\n", cpu_id); */

	if (ipi & MACPPC_IPI_FLUSH_FPU) {
		if (curcpu()->ci_fpulwp) {
			save_fpu(curcpu()->ci_fpulwp);
			if (curcpu()->ci_fpulwp)
				panic("cpuintr");
		}
	}
	if (ipi & MACPPC_IPI_HALT) {
		printf("halt{%d}\n", cpu_id);
		asm volatile ("mfmsr %0" : "=r"(msr));
		msr &= ~PSL_EE;
		msr |= PSL_POW;
		for (;;) {
			asm volatile ("sync; isync");
			asm volatile ("mtmsr %0" :: "r"(msr));
		}
	}
	s = splhigh();
	simple_lock(&ipilock);
	IPI[cpu_id] &= ~ipi;
	simple_unlock(&ipilock);
	splx(s);

	return 1;
}
#endif /* MULTIPROCESSOR */
