/*	$Id: clock.c,v 1.8 2003/02/22 18:08:34 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <machine/cpu.h>
#include <machine/mpc860dev.h>
#include <machine/sysconf.h>

#include <dev/clock_subr.h>

/*
 * Initially we assume TMBCLK is 4 MHz.
 */
static u_long ticks_per_sec = TIMEBASE_FREQ / 4;
static u_long ns_per_tick = 1000;
long ticks_per_intr;
static int clockinitted;

static volatile u_long lasttb;
volatile int tickspending;

#define DIFF19041970 2082844800

#define DS1307_ADDR	0xd0

struct ds1307_time {
	char sec;
	char min;
	char hour;
	char day;	/* Day of week (1-7, Sun = 1) */
	char date;
	char month;
	char year;
	char control;
};

void
read_rtc(dt)
	struct clock_ymdhms *dt;
{
	struct ds1307_time rtc;

	i2c_read(DS1307_ADDR, 0, sizeof rtc, &rtc);
	dt->dt_year = FROMBCD(rtc.year) + 2000;
	dt->dt_mon  = FROMBCD(rtc.month);
	dt->dt_day  = FROMBCD(rtc.date);
	dt->dt_wday = rtc.day - 1;
	dt->dt_hour = FROMBCD(rtc.hour);
	dt->dt_min  = FROMBCD(rtc.min);
	dt->dt_sec  = FROMBCD(rtc.sec);

	if (rtc.hour & 0x40) {		/* 12h/24h */
		int hour = FROMBCD(rtc.hour & 0x1f) % 12;

		if (rtc.hour & 0x20)	/* p.m.? */
			hour += 12;
		dt->dt_hour = hour;
	}
}

void
cpu_initclocks()
{
}

void
inittodr(base)
	time_t base;
{
	struct clock_ymdhms dt;
	int badbase = 0, deltat;

	if (base < 30 * SECYR) {
		printf("WARNING: preposterous time in file system\n");
		/* read the system clock anyway */
		badbase = 1;
	}

	read_rtc(&dt);
	clockinitted = 1;

	/* simple sanity checks */
	if (dt.dt_mon < 1 || dt.dt_mon > 12 ||
	    dt.dt_day < 1 || dt.dt_day > 31 ||
	    dt.dt_hour > 23 || dt.dt_min > 59 || dt.dt_sec > 59) {
		printf("WARNING: preposterous clock chip time\n");
		/*
		 * Believe the time in the file system for lack of
		 * anything better, resetting the TODR.
		 */
		time.tv_sec = base;
		if (!badbase)
			resettodr();
		return;
	}

	time.tv_sec = clock_ymdhms_to_secs(&dt) + rtc_offset * 60;

	if (!badbase) {
		deltat = time.tv_sec - base;
		if (deltat < 0)
			deltat = -deltat;
		if (deltat < 2 * SECDAY)
			return;
		printf("WARNING: clock %s %d days",
		    time.tv_sec < base ? "lost" : "gained", deltat / SECDAY);
	}

	printf(" -- CHECK AND RESET THE DATE!\n");
}

/*
 * Similar to the above
 */
void
resettodr()
{
	struct clock_ymdhms dt;
	struct ds1307_time rtc;

	if (clockinitted) {
		clock_secs_to_ymdhms(time.tv_sec - rtc_offset * 60, &dt);
		rtc.year  = TOBCD(dt.dt_year - 2000);
		rtc.month = TOBCD(dt.dt_mon);
		rtc.date  = TOBCD(dt.dt_day);
		rtc.day   = dt.dt_wday + 1;
		rtc.hour  = TOBCD(dt.dt_hour);
		rtc.min   = TOBCD(dt.dt_min);
		rtc.sec   = TOBCD(dt.dt_sec);
		/* set_rtc(&rtc); */
	}
}

void
decr_intr(frame)
	struct clockframe *frame;
{
	u_long tb;
	long tick;
	int nticks;
	int pri, msr;

	/* Check whether we are initialized. */
	if (!ticks_per_intr)
		return;

	/*
	 * Based on the actual time delay since the last decrementer reload,
	 * we arrange for earlier interrupt next time.
	 */
	asm ("mfdec %0" : "=r"(tick));
	for (nticks = 0; tick < 0; nticks++)
		tick += ticks_per_intr;
	asm volatile ("mtdec %0" :: "r"(tick));

	uvmexp.intrs++;
	intrcnt[CNT_CLOCK]++;

	pri = splclock();
	if (pri & (1 << SPL_CLOCK))
		tickspending += nticks;
	else {
		nticks += tickspending;
		tickspending = 0;

		/*
		 * lasttb is used during microtime. Set it to the virtual
		 * start of this tick interval.
		 */
		asm ("mftb %0" : "=r"(tb));
		lasttb = tb + tick - ticks_per_intr;

		/*
		 * Reenable interrupts
		 */
		asm volatile ("mfmsr %0; ori %0, %0, %1; mtmsr %0"
			      : "=r"(msr) : "K"(PSL_EE));
		
		/*
		 * Do standard timer interrupt stuff.
		 * Do softclock stuff only on the last iteration.
		 */
		frame->pri = pri | (1 << SIR_CLOCK);
		while (--nticks > 0)
			hardclock(frame);
		frame->pri = pri;
		hardclock(frame);

		/* Service software watchdog timer. */
		sswt();
	}
	splx(pri);
}

void
timebase_init()
{
	volatile struct mpc860dev *immr = get_immr();
	int msr, scratch;
	
	ticks_per_sec = TIMEBASE_FREQ / 4;

	asm volatile ("mfmsr %0; andi. %1,%0,%2; mtmsr %1"
		      : "=r"(msr), "=r"(scratch) : "K"((u_short)~PSL_EE));
	ns_per_tick = 1000000000 / ticks_per_sec;
	ticks_per_intr = ticks_per_sec / hz;
	asm volatile ("mftb %0" : "=r"(lasttb));
	asm volatile ("mtdec %0" :: "r"(ticks_per_intr));
	asm volatile ("mtmsr %0" :: "r"(msr));

	/* Start timebase counter. */
	immr->siu_tbscr = 0x0001;
}

static inline u_quad_t
mftb()
{
	u_long scratch;
	u_quad_t tb;
	
	asm ("1: mftbu %0; mftb %0+1; mftbu %1; cmpw 0,%0,%1; bne 1b"
	    : "=r"(tb), "=r"(scratch));
	return tb;
}

/*
 * Fill in *tvp with current time with microsecond resolution.
 */
void
microtime(tvp)
	struct timeval *tvp;
{
	u_long tb;
	u_long ticks;
	int msr, scratch;
	
	asm volatile ("mfmsr %0; andi. %1,%0,%2; mtmsr %1"
		      : "=r"(msr), "=r"(scratch) : "K"((u_short)~PSL_EE));
	asm ("mftb %0" : "=r"(tb));
	ticks = (tb - lasttb) * ns_per_tick;
	*tvp = time;
	asm volatile ("mtmsr %0" :: "r"(msr));
	ticks /= 1000;
	tvp->tv_usec += ticks;
	while (tvp->tv_usec >= 1000000) {
		tvp->tv_usec -= 1000000;
		tvp->tv_sec++;
	}
}

/*
 * Wait for about n microseconds (at least!).
 */
void
delay(n)
	unsigned n;
{
	u_quad_t tb;
	u_long tbh, tbl, scratch;

	/* Service watch dog timer before long wait while boot sequence. */
	if (/* n > 100000 && */ cold)
		sswt();
	
	tb = mftb();
	tb += (n * 1000 + ns_per_tick - 1) / ns_per_tick;
	tbh = tb >> 32;
	tbl = tb;
	asm volatile ("1: mftbu %0; cmplw %0,%1; blt 1b; bgt 2f;"
		      "mftb %0; cmplw %0,%2; blt 1b; 2:"
		      : "=r"(scratch) : "r"(tbh), "r"(tbl));
}

/*
 * Nothing to do.
 */
void
setstatclockrate(arg)
	int arg;
{
	/* Do nothing */
}
