/*	$Id: hfs_lookup.c,v 1.21 2004/01/09 11:05:37 tsubai Exp $	*/

/*-
 * Copyright (c) 2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <fs/hfs/hfs.h>

int
hfs_lookup(v)
	void *v;
{
	struct vop_lookup_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
	} */ *ap = v;
	struct vnode *vdp = ap->a_dvp;		/* searched directory */
	struct vnode **vpp = ap->a_vpp;		/* place to store result */
	struct componentname *cnp = ap->a_cnp;
	struct ucred *cred = cnp->cn_cred;
	struct hfsnode *hdp = VTOH(vdp);
	int error, nocache = 0;
	int nameiop = cnp->cn_nameiop;
	int flags = cnp->cn_flags;
	int last = flags & ISLASTCN;
	int lockp = flags & LOCKPARENT;
	int wantp = flags & (LOCKPARENT | WANTPARENT);
	const char *name = cnp->cn_nameptr;
	u_int nlen = cnp->cn_namelen;
	struct hfsmount *hmp;
	struct buf *bp;
	struct vnode *tdp;
	u_int node;

	*vpp = NULL;
	cnp->cn_flags &= ~PDIRUNLOCK;

	/*
	 * Check accessiblity of directory.
	 */
	if ((error = VOP_ACCESS(vdp, VEXEC, cred, cnp->cn_proc)) != 0)
		return error;

	if ((flags & ISLASTCN) && (vdp->v_mount->mnt_flag & MNT_RDONLY) &&
	    (cnp->cn_nameiop == DELETE || cnp->cn_nameiop == RENAME))
		return EROFS;

	/*
         * We now have a segment name to search for, and a directory to search.
	 *
	 * Before tediously performing a linear scan of the directory,
	 * check the name cache to see if the directory/name pair
	 * we are looking for is known already.
	 */
	if ((error = cache_lookup(vdp, vpp, cnp)) >= 0)
		return error;

	/*
	 * Fake a '.'
	 */
	if (nlen == 1 && name[0] == '.') {
		/* see special cases in prologue. */
		tdp = vdp;
		goto found;
	}

	/*
	 * Fake a "..".  We must get the target inode before unlocking
	 * the directory to insure that the inode will not be removed
	 * before we get it.  We prevent deadlock by always fetching
	 * inodes from the root, moving down the directory tree. Thus
	 * when following backward pointers ".." we must unlock the
	 * parent directory before getting the requested directory.
	 * There is a potential race condition here if both the current
	 * and parent directories are removed before the VFS_VGET for the
	 * inode associated with ".." returns.  We hope that this occurs
	 * infrequently since we cannot avoid this race condition without
	 * implementing a sophisticated deadlock detection algorithm.
	 * Note also that this simple deadlock detection scheme will not
	 * work if the file system has any hard links other than ".."
	 * that point backwards in the directory structure.
	 */
	if (flags & ISDOTDOT) {
		VOP_UNLOCK(vdp, 0);	/* race to get the inode */
		cnp->cn_flags |= PDIRUNLOCK;
		error = VFS_VGET(vdp->v_mount, hdp->parent, &tdp);
		if (error) {
			if (vn_lock(vdp, LK_EXCLUSIVE | LK_RETRY) == 0)
				cnp->cn_flags &= ~PDIRUNLOCK;
			return error;
		}
		if (last && lockp) {
			if ((error = vn_lock(vdp, LK_EXCLUSIVE)) != 0) {
				vput(tdp);
				return error;
			}
			cnp->cn_flags &= ~PDIRUNLOCK;
		}
		goto found_lockdone;
	}

	hmp = hdp->hmp;
	node = hmp->flnode;		/* ǽ leaf node */
	while (node) {
		int i, len, type, id;
		u_int nrec, recoff, dataoff;
		u_char *pnam;

		bp = read_catalog(hmp, node);
		if (bp == NULL)
			return EINVAL;
		nrec = GET16B(bp, ndNRecs);

		for (i = 0; i < nrec; i++) {
			recoff = GET16B(bp, NODESIZE - 2 - i * 2);
			len = GET8B(bp, recoff + ckrKeyLen);
			dataoff = (recoff + len + 2) & ~1;

			type = GET8B(bp, dataoff + cdrType);
			switch (type) {

			case 1: /* dir */
				id = GET32B(bp, dataoff + dirDirID);
				break;

			case 2: /* file */
				id = GET32B(bp, dataoff + filFlNum);
				break;

			default:
				continue;
			}

			/* File name in Pascal string. */
			pnam = bp->b_data + recoff + ckrCName;
			if (nlen == *pnam && bcmp(name, pnam + 1, nlen) == 0) {
				brelse(bp);
				error = VFS_VGET(vdp->v_mount, id, &tdp);
				if (error)
					return error;
				goto found;
			}
		}
		node = GET32B(bp, ndFLink);
		brelse(bp);
	}
	/* ޤǤǤ٤Ƥ record  */

	/*
	 * not found
	 */
	if ((nameiop == CREATE || nameiop == RENAME) && last) {

		if (vdp->v_mount->mnt_flag & MNT_RDONLY)
			return EROFS;

		if ((error = VOP_ACCESS(vdp, VWRITE, cred, cnp->cn_proc)) != 0)
			return error;
		if (lockp == 0) {
			VOP_UNLOCK(vdp, 0);
			cnp->cn_flags |= PDIRUNLOCK;
		}
		cnp->cn_nameiop |= SAVENAME;
		return EJUSTRETURN;
	}
	if ((cnp->cn_flags & MAKEENTRY) && nameiop != CREATE)
		cache_enter(vdp, NULL, cnp);
	return ENOENT;

found:
	if (nameiop == DELETE && last) {
		error = VOP_ACCESS(vdp, VWRITE, cred, cnp->cn_proc);
		if (error) {
			if (vdp != tdp)
				vput(tdp);
			return error;
		}
		nocache = 1;
	}
	if (nameiop == RENAME && wantp && last) {
		if (vdp == tdp)
			return EISDIR;
		error = VOP_ACCESS(vdp, VWRITE, cred, cnp->cn_proc);
		if (error) {
			vput(tdp);
			return error;
		}
		cnp->cn_flags |= SAVENAME;
		nocache = 1;
	}
	if (vdp == tdp)
		VREF(vdp);
	else if (lockp == 0 || last == 0) {
		VOP_UNLOCK(vdp, 0);
		cnp->cn_flags |= PDIRUNLOCK;
	}

found_lockdone:
	*vpp = tdp;
	if ((cnp->cn_flags & MAKEENTRY) && nocache == 0)
		cache_enter(vdp, *vpp, cnp);

	return 0;
}
