/*	$Id: hfs_vfsops.c,v 1.27 2004/01/08 11:26:58 tsubai Exp $	*/

/*-
 * Copyright (c) 2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/malloc.h>
#include <sys/namei.h>
#include <sys/sysctl.h>
#include <sys/vnode.h>
#include <miscfs/specfs/specdev.h> /* XXX */
#include <fs/hfs/hfs.h>

int hfs_mount __P((struct mount *, const char *, void *, struct nameidata *,
		      struct proc *));
int hfs_start __P((struct mount *, int, struct proc *));
int hfs_unmount __P((struct mount *, int, struct proc *));
int hfs_root __P((struct mount *, struct vnode **));
int hfs_quotactl __P((struct mount *, int, uid_t, caddr_t, struct proc *));
int hfs_statfs __P((struct mount *, struct statfs *, struct proc *));
int hfs_sync __P((struct mount *, int, struct ucred *, struct proc *));
int hfs_vget __P((struct mount *, ino_t, struct vnode **));
int hfs_fhtovp __P((struct mount *, struct fid *, struct vnode **));
int hfs_vptofh __P((struct vnode *, struct fid *));
void hfs_init __P((void));
void hfs_done __P((void));
int hfs_checkexp __P((struct mount *, struct mbuf *, int *, struct ucred **));

int hfs_mountfs __P((struct vnode *, struct mount *, struct proc *));

struct pool hfs_node_pool;

MALLOC_DEFINE(M_HFSMNT, "hfs mount", "hfs mount structures");
MALLOC_DEFINE(M_HFSBITMAP, "hfs bitmap", "hfs bitmap");

struct genfs_ops hfs_genfsops = {
	genfs_size,
};

int
hfs_mount(mp, path, data, ndp, p)
	struct mount *mp;
	const char *path;
	void *data;
	struct nameidata *ndp;
	struct proc *p;
{
	struct vnode *devvp;
	struct hfs_args args;
	struct hfsmount *hmp;
	int error;
	mode_t accessmode;

	if (mp->mnt_flag & MNT_GETARGS) {
		hmp = VFSTOHFS(mp);
		if (hmp == NULL)
			return EIO;
		args.fspec = NULL;
		args.uid = hmp->uid;
		args.gid = hmp->gid;
		args.mask = hmp->mask;
		vfs_showexport(mp, &args.export, &hmp->export);
		return copyout(&args, data, sizeof args);
	}
	error = copyin(data, &args, sizeof(struct hfs_args));
	if (error)
		return error;

	if ((mp->mnt_flag & MNT_RDONLY) == 0)
		return EROFS;
	/*
	 * If updating, check whether changing from read-only to
	 * read/write; if there is no device name, that's all we do.
	 */
	if (mp->mnt_flag & MNT_UPDATE) {
		hmp = VFSTOHFS(mp);
		if (args.fspec == 0)
			return vfs_export(mp, &hmp->export, &args.export);
	}

	/*
	 * Not an update, or updating the name: look up the name
	 * and verify that it refers to a sensible block device.
	 */
	NDINIT(ndp, LOOKUP, FOLLOW, UIO_USERSPACE, args.fspec, p);
	if ((error = namei(ndp)) != 0)
		return error;
	devvp = ndp->ni_vp;

	if (devvp->v_type != VBLK) {
		vrele(devvp);
		return ENOTBLK;
	}
	if (bdevsw_lookup(devvp->v_rdev) == NULL) {
		vrele(devvp);
		return ENXIO;
	}

	/*
	 * If mount by non-root, then verify that user has necessary
	 * permissions on the device.
	 */
	if (p->p_ucred->cr_uid != 0) {
		accessmode = VREAD;
		if ((mp->mnt_flag & MNT_RDONLY) == 0)
			accessmode |= VWRITE;
		vn_lock(devvp, LK_EXCLUSIVE | LK_RETRY);
		error = VOP_ACCESS(devvp, accessmode, p->p_ucred, p);
		VOP_UNLOCK(devvp, 0);
		if (error) {
			vput(devvp);
			return error;
		}
	}

	/* XXX MNT_UPDATE */
	if ((error = hfs_mountfs(devvp, mp, p)) != 0) {
		vrele(devvp);
		return error;
	}
	hmp = VFSTOHFS(mp);
	hmp->uid = args.uid;
	hmp->gid = args.gid;
	hmp->mask = args.mask;
	return set_statfs_info(path, UIO_USERSPACE, args.fspec, UIO_USERSPACE,
	    mp, p);
}

int
hfs_mountfs(devvp, mp, p)
	struct vnode *devvp;
	struct mount *mp;
	struct proc *p;
{
	struct hfsmount *hmp;
	struct buf *bp;
	struct vnode *rvp;
	int error;

	hmp = NULL;

	/*
	 * Disallow multiple mounts of the same device.
	 * Disallow mounting of a device that is currently in use
	 * (except for root, which might share swap device for miniroot).
	 * Flush out any old buffers remaining from a previous use.
	 */
	if ((error = vfs_mountedon(devvp)) != 0)
		return error;
	if (vcount(devvp) > 1 && devvp != rootvp)
		return EBUSY;
	if ((error = vinvalbuf(devvp, V_SAVE, p->p_ucred, p, 0, 0)) != 0)
		return error;

	/*
	 * Open device and read MDB (Master Directory Block).
	 */
	if ((error = VOP_OPEN(devvp, FREAD, NOCRED, p)) != 0)
		return error;

	hmp = malloc(sizeof(struct hfsmount), M_HFSMNT, M_WAITOK);
	bzero(hmp, sizeof(struct hfsmount));
	hmp->mp = mp;
	hmp->devvp = devvp;
	hmp->bsize = 512;	/* XXX HFS */

	/* Read MDB.  XXX DEV_BSIZE */
	if ((error = bread(devvp, 2, hmp->bsize, NOCRED, &bp)) != 0) {
		brelse(bp);
		goto fail;
	}
	hmp->magic = GET16B(bp, drSigWord);
	hmp->ablkoff = GET16B(bp, drAlBlSt);
	hmp->ablksize = GET32B(bp, drAlBlkSiz);
	hmp->catsize = GET32B(bp, drCTFlSize);	// ?

	hmp->catABN[0] = GET16B(bp, drCTExtRec + 0);	// ?
	hmp->catABlen[0] = GET16B(bp, drCTExtRec + 2);
	hmp->catABN[1] = GET16B(bp, drCTExtRec + 4);
	hmp->catABlen[1] = GET16B(bp, drCTExtRec + 6);
	hmp->catABN[2] = GET16B(bp, drCTExtRec + 8);
	hmp->catABlen[3] = GET16B(bp, drCTExtRec + 10);

	hmp->freeblks = GET16B(bp, drFreeBks) * (hmp->ablksize / 512);
	hmp->numblks = GET16B(bp, drNmAlBlks) * (hmp->ablksize / 512);

	brelse(bp);

	/* basic sanity checks */
	if (hmp->magic != HFS_MAGIC) {
		error = EINVAL;
		goto fail;
	}

#if 0
printf("allocation block starts at 0x%x\n", hmp->ablkoff * 512);
printf("catalog file size = 0x%x\n", hmp->catsize);
printf("catalog file[0] 0x%x 0x%x\n", hmp->catABN[0], hmp->catABlen[0]);
printf("catalog file[1] 0x%x 0x%x\n", hmp->catABN[1], hmp->catABlen[1]);
printf("catalog file[2] 0x%x 0x%x\n", hmp->catABN[2], hmp->catABlen[2]);
#endif

	/*
	 * Read the B*-tree header.
	 */
	bp = read_catalog(hmp, 0);
	if (bp == NULL) {
		error = EINVAL;
		goto fail;
	}

#ifdef DIAGNOSTIC
	if (GET16B(bp, 510) != 0x0e) {			/* XXX HFS */
		brelse(bp);
		printf("hfs: recoff != 0x0e\n");
		error = EINVAL;
		goto fail;
	}
#endif
	hmp->flnode = GET32B(bp, 0x0e + bthFNode);	/* XXX HFS */

#if 0
printf("B*-tree header:\n");
printf("bthRoot = 0x%x, ", GET32B(bp, 0x0e + bthRoot));
printf("bthKeyLen = %d, ", GET16B(bp, 0x0e + bthKeyLen));
printf("bthFNode = 0x%x, ", GET32B(bp, 0x0e + bthFNode));
printf("bthLNode = 0x%x, ", GET32B(bp, 0x0e + bthLNode));
printf("bthNNodes = 0x%x\n", GET32B(bp, 0x0e + bthNNodes));
#endif

	brelse(bp);

	mp->mnt_data = hmp;
	mp->mnt_stat.f_fsid.val[0] = (long)devvp->v_rdev;
	mp->mnt_stat.f_fsid.val[1] = makefstype(MOUNT_HFS);
	mp->mnt_fs_bshift = ffs(hmp->bsize) - 1;
	mp->mnt_dev_bshift = DEV_BSHIFT;	/* XXX */
	mp->mnt_flag |= MNT_LOCAL;

	/*
	 * Get the root hfsnode for sanity check.
	 */
	if ((error = VFS_ROOT(mp, &rvp)) != 0)
		goto fail;
	vput(rvp);

	if (mp->mnt_flag & MNT_RDONLY)
		return 0;

	hmp->bitmap = malloc((hmp->ablkoff - 3) * 512, M_HFSBITMAP, M_WAITOK);
#ifdef DIAGNOSTIC
	if ((hmp->ablkoff - 3) * 4096 < hmp->numblks / (hmp->ablksize / 512))
		panic("hfs: bad bitmap");
#endif

	error = bread(devvp, 3, (hmp->ablkoff - 3) * 512, NOCRED, &bp);
	if (error) {
		brelse(bp);
		goto fail;
	}
	bcopy(bp->b_data, hmp->bitmap, (hmp->ablkoff - 3) * 512);
	brelse(bp);
	return 0;

fail:
	vn_lock(devvp, LK_EXCLUSIVE | LK_RETRY);
	VOP_CLOSE(devvp, FREAD, NOCRED, p);
	VOP_UNLOCK(devvp, 0);
	if (hmp && hmp->bitmap)
		free(hmp->bitmap, M_HFSBITMAP);
	if (hmp)
		free(hmp, M_HFSMNT);
	return error;
}

int
hfs_start(mp, flags, p)
	struct mount *mp;
	int flags;
	struct proc *p;
{

	return 0;
}

int
hfs_unmount(mp, mntflags, p)
	struct mount *mp;
	int mntflags;
	struct proc *p;
{
	struct hfsmount *hmp;
	int error, flags;

	flags = 0;
	if (mntflags & MNT_FORCE)
		flags |= FORCECLOSE;
	if ((error = vflush(mp, NULLVP, flags)) != 0)
		return error;
	hmp = VFSTOHFS(mp);
	if (hmp->devvp->v_type != VBAD)
		hmp->devvp->v_specmountpoint = NULL;
	vn_lock(hmp->devvp, LK_EXCLUSIVE | LK_RETRY);
	error = VOP_CLOSE(hmp->devvp, FREAD, NOCRED, p);
	vput(hmp->devvp);
	if (hmp->bitmap)
		free(hmp->bitmap, M_HFSBITMAP);
	free(hmp, M_HFSMNT);
	mp->mnt_data = NULL;
	mp->mnt_flag &= ~MNT_LOCAL;
	return error;
}

int
hfs_root(mp, vpp)
	struct mount *mp;
	struct vnode **vpp;
{
	struct vnode *nvp;
	int error;

	/* CNID2 is root directory. */
	if ((error = VFS_VGET(mp, 2, &nvp)) != 0)
		return error;
	*vpp = nvp;
	return 0;
}

int
hfs_quotactl(mp, cmds, uid, arg, p)
	struct mount *mp;
	int cmds;
	uid_t uid;
	caddr_t arg;
	struct proc *p;
{
	return EOPNOTSUPP;
}

int
hfs_statfs(mp, sbp, p)
	struct mount *mp;
	struct statfs *sbp;
	struct proc *p;
{
	struct hfsmount *hmp;

	hmp = VFSTOHFS(mp);

	sbp->f_type = 0;
	sbp->f_bsize = hmp->bsize;	// 512?
	sbp->f_iosize = hmp->bsize;	// allocation block size?
	sbp->f_blocks = hmp->numblks;
	sbp->f_bfree = hmp->freeblks;
	sbp->f_bavail = sbp->f_bfree;
	sbp->f_files = 0;		/* XXX */
	sbp->f_ffree = 0;		/* XXX */
	copy_statfs_info(sbp, mp);
	return 0;
}

int
hfs_sync(mp, waitfor, uc, p)
	struct mount *mp;
	int waitfor;
	struct ucred *uc;
	struct proc *p;
{
	return 0;
}

#define DIFF19041970 2082844800

/*
 * lookup an hfsnode, check mount's hash table if not found, create
 * return locked and referenced al la vget(vp, 1);
 */
int
hfs_vget(mp, an, vpp)
	struct mount *mp;
	ino_t an;
	struct vnode **vpp;
{
	struct hfsmount *hmp;
	struct vnode *vp;
	struct hfsnode *hp;
	struct buf *bp;
	int error = 0;

	hmp = VFSTOHFS(mp);

//printf("hfs_vget ino = 0x%x\n", an);
	*vpp = NULL;

	u_int node = hmp->flnode;		/* ǽ leaf node */
	int i, len, type, id;
	u_int nrec, recoff, dataoff;

nextleaf:
	bp = read_catalog(hmp, node);
	if (bp == NULL)
		return EINVAL;

	nrec = GET16B(bp, ndNRecs);

	for (i = 0; i < nrec; i++) {
		recoff = GET16B(bp, NODESIZE - 2 - i * 2);
		len = GET8B(bp, recoff + ckrKeyLen);
		dataoff = (recoff + len + 2) & ~1;

		type = GET8B(bp, dataoff + cdrType);
		switch (type) {

		case 1: /* dir */
			id = GET32B(bp, dataoff + dirDirID);
			if (id == an) goto found;
			break;

		case 2: /* file */
			id = GET32B(bp, dataoff + filFlNum);
			if (id == an) goto found;
			break;
		}
	}
	node = GET32B(bp, ndFLink);
	brelse(bp);
	if (node)
		goto nextleaf;

printf("hfs_vget ENOENT\n");
	return ENOENT;		/* XXX ? */

found:
	error = getnewvnode(VT_HFS, mp, hfs_vnodeop_p, &vp);
	if (error) {
		brelse(bp);
		return error;
	}

	/*
	 * setup, insert in hash, and lock before io.
	 */
	vp->v_data = hp = pool_get(&hfs_node_pool, PR_WAITOK);
	bzero(hp, sizeof(struct hfsnode));
	hp->vp = vp;
	hp->hmp = hmp;
	lockmgr(&hp->vp->v_lock, LK_EXCLUSIVE, (struct simplelock *)0);

	hp->parent = GET32B(bp, recoff + ckrParID);
	if (hp->parent == 1)
		hp->parent = 2;

	/*
	 * get type and fill rest in based on that.
	 */
	switch (type) {
	case 1: /* dir (2-73) */
		vp->v_type = VDIR;
		vp->v_flag |= VROOT;
		hp->mtime = GET32B(bp, dataoff + dirMdDat) - DIFF19041970;
		hp->ctime = GET32B(bp, dataoff + dirCrDat) - DIFF19041970;
		hp->cnid = GET32B(bp, dataoff + dirDirID);
		break;
	case 2: /* file (2-73) */
		vp->v_type = VREG;
		hp->fsize = GET32B(bp, dataoff + filLgLen);
		hp->psize = GET32B(bp, dataoff + filPyLen);
		hp->mtime = GET32B(bp, dataoff + filMdDat) - DIFF19041970;
		hp->ctime = GET32B(bp, dataoff + filCrDat) - DIFF19041970;
		hp->cnid = GET32B(bp, dataoff + filFlNum);

		hp->abnum[0] = GET16B(bp, dataoff + filExtRec + 0);
		hp->ablen[0] = GET16B(bp, dataoff + filExtRec + 2);
		hp->abnum[1] = GET16B(bp, dataoff + filExtRec + 4);
		hp->ablen[1] = GET16B(bp, dataoff + filExtRec + 6);
		hp->abnum[2] = GET16B(bp, dataoff + filExtRec + 8);
		hp->ablen[2] = GET16B(bp, dataoff + filExtRec + 10);
		break;
	default:
		brelse(bp);
		vput(vp);
		return EINVAL;
	}

	hp->uid = hmp->uid;
	hp->gid = hmp->gid;

	genfs_node_init(vp, &hfs_genfsops);
	*vpp = vp;
	brelse(bp);
	vp->v_size = hp->fsize;
	return 0;
}

struct ifid {
	u_short ifid_len;
	u_short ifid_pad;
	int ifid_ino;
};

int
hfs_fhtovp(mp, fhp, vpp)
	struct mount *mp;
	struct fid *fhp;
	struct vnode **vpp;
{
	struct ifid *ifhp = (struct ifid *)fhp;
	struct vnode *nvp;
	int error;

	if ((error = VFS_VGET(mp, ifhp->ifid_ino, &nvp)) != 0) {
		*vpp = NULLVP;
		return error;
	}
	*vpp = nvp;
	return 0;
}

int
hfs_vptofh(vp, fhp)
	struct vnode *vp;
	struct fid *fhp;
{
	struct hfsnode *hp = VTOH(vp);
	struct ifid *ifhp;

	ifhp = (struct ifid *)fhp;
	ifhp->ifid_len = sizeof(struct ifid);
	ifhp->ifid_ino = hp->cnid;

	return 0;
}

void
hfs_init()
{
	pool_init(&hfs_node_pool, sizeof(struct hfsnode), 0, 0, 0,
	    "hfsndpl", &pool_allocator_nointr);
}

void
hfs_done()
{
	pool_destroy(&hfs_node_pool);
}

int
hfs_checkexp(mp, nam, exflagsp, credanonp)
	struct mount *mp;
	struct mbuf *nam;
	int *exflagsp;
	struct ucred **credanonp;
{
	struct hfsmount *hmp = VFSTOHFS(mp);
	struct netcred *np;

	np = vfs_export_lookup(mp, &hmp->export, nam);
	if (np == NULL)
		return EACCES;
	*exflagsp = np->netc_exflags;
	*credanonp = &np->netc_anon;
	return 0;
}

#if 0
SYSCTL_SETUP(sysctl_vfs_hfs_setup, "sysctl vfs.hfs subtree setup")
{

	sysctl_createv(SYSCTL_PERMANENT,
		       CTLTYPE_NODE, "vfs", NULL,
		       NULL, 0, NULL, 0,
		       CTL_VFS, CTL_EOL);
	sysctl_createv(SYSCTL_PERMANENT,
		       CTLTYPE_NODE, "hfs", NULL,
		       NULL, 0, NULL, 0,
		       CTL_VFS, 21, CTL_EOL);
	/*
	 * XXX the "21" above could be dynamic, thereby eliminating
	 * one more instance of the "number to vfs" mapping problem,
	 * but "21" is the order as taken from sys/mount.h
	 */
}
#endif

/*
 * vfs generic function call table
 */

extern const struct vnodeopv_desc hfs_vnodeop_opv_desc;

const struct vnodeopv_desc *hfs_vnodeopv_descs[] = {
	&hfs_vnodeop_opv_desc,
	NULL,
};

struct vfsops hfs_vfsops = {
	MOUNT_HFS,
	hfs_mount,
	hfs_start,
	hfs_unmount,
	hfs_root,
	hfs_quotactl,
	hfs_statfs,
	hfs_sync,
	hfs_vget,
	hfs_fhtovp,
	hfs_vptofh,
	hfs_init,
	NULL,
	hfs_done,
	NULL,
	NULL,				/* vfs_mountroot */
	hfs_checkexp,
	hfs_vnodeopv_descs,
};
